# -*- coding: utf-8 -*-
#
# pygenda_gui.py
# Top-level GUI code and shared elements (e.g. soft buttons, dialogs)
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib, Gio

from datetime import date as dt_date, time as dt_time, datetime as dt_datetime, timedelta
from dateutil import rrule as du_rrule
from dateutil.relativedelta import relativedelta
from icalendar import Event as iEvent
from importlib import import_module
from os import path as ospath
from sys import stderr
import signal
import ctypes
from typing import Optional, Tuple

# for internationalisation/localisation
import locale
_ = locale.gettext
from calendar import day_name, monthrange

# pygenda components
from .pygenda_config import Config
from .pygenda_calendar import Calendar, RepeatInfo
from .pygenda_widgets import WidgetDate, WidgetTime, WidgetDuration
from .pygenda_entryinfo import EntryInfo
from .pygenda_util import guess_date_ord_from_locale,guess_date_sep_from_locale,guess_time_sep_from_locale, guess_date_fmt_text_from_locale, datetime_to_date
from .pygenda_version import __version__


# Singleton class for top-level GUI control
class GUI:
	_GLADE_FILE = '{:s}/pygenda.glade'.format(ospath.dirname(__file__))
	_CSS_FILE_APP = '{:s}/css/pygenda.css'.format(ospath.dirname(__file__))
	_CSS_FILE_USER = GLib.get_user_config_dir() + '/pygenda/pygenda.css'
	_LOCALE_DIR = '{:s}/locale/'.format(ospath.dirname(__file__))
	_VIEWS = ('Week','Year') # Order gives order in menus
	SPINBUTTON_INC_KEY = (Gdk.KEY_plus,Gdk.KEY_greater)
	SPINBUTTON_DEC_KEY = (Gdk.KEY_minus,Gdk.KEY_less)
	STYLE_ERR = 'dialog_error'

	cursor_date = dt_date.today()
	cursor_idx_in_date = 0 # cursor index within date
	today_toggle_date = None
	today_toggle_idx = 0

	Config.set_defaults('global',{
		'language': '', # use OS language
		'hide_titlebar_when_maximized': False,
		'date_ord': guess_date_ord_from_locale(),
		'date_sep': guess_date_sep_from_locale(),
		'time_sep': guess_time_sep_from_locale(),
		'date_fmt_text': guess_date_fmt_text_from_locale(),
		'date_fmt_text_noyear': '', # construct from date_fmt_text
		'date_fmt_textabb': '', # construct from date_fmt_text
		'date_fmt_textabb_noyear': '', # construct from date_fmtabb_text
		'24hr': False,
		'start_week_day': 0, # 0 = Monday, 6 = Sunday
		'tab_elts_datetime': False,
		})
	Config.set_defaults('startup',{
		'maximize': False,
		'fullscreen': False,
		'view': False,
		'softbutton_display': '',
		})

	# Constructor
	@classmethod
	def init(cls) -> None:
		# First stage initialisation to bring up the UI.
		# See init_stage2() below for init done after gtk_main loop started.

		# First set the locale, so UI language (e.g. in menu) is correct
		cls._init_locale()

		# Construct GUI from GTK Builder XML glade file
		cls._builder = Gtk.Builder()
		cls._builder.add_from_file(cls._GLADE_FILE)

		cls._window = cls._builder.get_object('window_main')
		if (not cls._window): # Sanity check
			raise NameError('Main window not found')
		cls._window.set_default_icon_name('x-office-calendar')

		cls._window.set_hide_titlebar_when_maximized(Config.get_bool('global','hide_titlebar_when_maximized'))
		if Config.get_bool('startup','maximize'):
			cls._window.maximize()
		cls._is_fullscreen = Config.get_bool('startup','fullscreen')
		if cls._is_fullscreen:
			cls._window.fullscreen()

		# Handle SIGINT (e.g. from ctrl+C) etc.
		GLib.unix_signal_add(GLib.PRIORITY_DEFAULT_IDLE, signal.SIGINT, cls.exit)
		GLib.unix_signal_add(GLib.PRIORITY_DEFAULT_IDLE, signal.SIGTERM, cls.exit)

		# Connect signals now, so clicking on [X] in window exits application
		HANDLERS = {
			'window_main delete': cls.exit,
			'menuitem_quit': cls.exit,
			'menuitem_cut': cls.cut_request,
			'menuitem_copy': cls.copy_request,
			'menuitem_paste': cls.paste_request,
			'menuitem_newentry': cls.event_newentry,
			'menuitem_edittime': cls.event_edittime,
			'menuitem_editrepeats': cls.event_editrepeats,
			'menuitem_editalarm': cls.event_editalarm,
			'menuitem_editdetails': cls.event_editdetails,
			'menuitem_stat_none': lambda a: cls.event_stat_toggle(None),
			'menuitem_stat_confirmed': lambda a: cls.event_stat_toggle('CONFIRMED'),
			'menuitem_stat_canceled': lambda a: cls.event_stat_toggle('CANCELLED'),
			'menuitem_stat_tentative': lambda a: cls.event_stat_toggle('TENTATIVE'),
			'menuitem_deleteentry': cls.delete_request,
			'menuitem_switchview': cls.switch_view,
			'menuitem_goto': cls.dialog_goto,
			'menuitem_fullscreen': cls.toggle_fullscreen,
			'menuitem_about': cls.dialog_about,
			'button0_clicked': cls.event_newentry,
			'button1_clicked': cls.switch_view,
			'button2_clicked': cls.dialog_goto,
			'button3_clicked': cls.debug, # zoom, to be decided/implemented
			'exceptions_modify': EntryDialogController.dialog_repeat_exceptions
			}
		cls._builder.connect_signals(HANDLERS)

		cls._box_view_cont = cls._builder.get_object('box_view_cont')
		if (not cls._box_view_cont): # Sanity check
			raise NameError('View container not found')

		# Add a spinner while the view is still loading
		cls._loading_indicator = Gtk.Box()
		cls._loading_indicator.set_name('view_loading')
		spinner = Gtk.Spinner.new()
		cls._loading_indicator.set_center_widget(spinner)
		cls._box_view_cont.pack_start(cls._loading_indicator, True, True, 0)
		spinner.start()
		cls._loading_indicator.show_all()

		# Setup CSS provider(s) now so "loading" notice is styled
		css_prov = Gtk.CssProvider()
		css_prov.load_from_path(cls._CSS_FILE_APP)
		Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), css_prov, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
		try:
			css_prov_u = Gtk.CssProvider()
			css_prov_u.load_from_path(cls._CSS_FILE_USER)
			Gtk.StyleContext.add_provider_for_screen(Gdk.Screen.get_default(), css_prov_u, Gtk.STYLE_PROVIDER_PRIORITY_USER)
		except:
			pass

		# Set position/display of softbuttons before showing loading indicator
		cfig_sbut = Config.get('startup','softbutton_display').lower()
		bbut = cls._builder.get_object('box_buttons')
		if cfig_sbut == 'hide':
			bbut.hide()
		elif cfig_sbut == 'left':
			cls._box_view_cont.reorder_child(bbut,0)
		else:
			cls._box_view_cont.reorder_child(bbut,1)

		# Delay further initialisation so we can display GUI/window early
		cls._window.show()
		GLib.idle_add(cls.init_stage2)


	@staticmethod
	def init_cal(task:Gio.Task, src_obj, t_data, cancel:Gio.Cancellable)->None:
		# Wrapper around Calendar.init(), so we can call it in a GTask thread.
		Calendar.init()
		task.return_int(0)

	@classmethod
	def init_cal_callback(cls, data, task:Gio.Task) -> None:
		# Callback called to signal that init_cal() has finished.
		cls._starting_cal = False

	@staticmethod
	def init_cal_iosc(job:Gio.IOSchedulerJob, cancel:Gio.Cancellable, u_data) -> bool:
		# Wrapper around Calendar.init, so we can call it in an IOScheduler.
		# Fallback for Gio version < 2.36 (hence needed for Gemini).
		try:
			Calendar.init()
		finally:
			GUI._starting_cal = False
		return False # indicates task complete


	@classmethod
	def init_stage2(cls) -> None:
		# Second-stage initialisation.
		# This is run in the gtk_main thread after basic UI is displayed.

		# First, initialise calendar connector.
		# Do this outside main thread, so UI remains active/responsive.
		ci_cancel = Gio.Cancellable.new()
		cls._starting_cal = True
		if 'run_in_thread' in dir(Gio.Task):
			# Preferred way to run in separate thread.
			task = Gio.Task.new(cancellable=ci_cancel, callback=cls.init_cal_callback)
			task.run_in_thread(cls.init_cal)
		else:
			# Deprecated, but needed to run on Gemini.
			Gio.io_scheduler_push_job(cls.init_cal_iosc, None, GLib.PRIORITY_DEFAULT, ci_cancel)

		# Initialise some things in main thread while calendar loading
		cls._init_clipboard()
		cls._init_date_format()
		EntryDialogController.init()

		# If date set from command line, jump there now
		if Config.date:
			GUI.cursor_date = Config.date

		# Wait for calendar to finish initialising before doing views
		while cls._starting_cal:
			if Gtk.main_iteration_do(False):
				# True => main_quit() has been called
				ci_cancel.cancel()
				return
			GLib.usleep(1000) # microsecs
		del(cls._starting_cal) # Flag no longer needed

		# Get rid of loading indicator, but store position for view placement
		view_pos = cls._box_view_cont.child_get_property(cls._loading_indicator, 'position')
		cls._box_view_cont.remove(cls._loading_indicator)
		del(cls._loading_indicator) # should be only remaining reference

		# Check that calendar initialised/connected properly
		if not hasattr(Calendar, 'calConnector') or not Calendar.calConnector:
			print('Error: Calendar could not be initialized', file=stderr)
			Gtk.main_quit()
			return

		cls._init_views()

		cls._view_idx = 0 # take first view as default
		vw = Config.get('startup','view')
		if vw:
			for ii in range(len(GUI._VIEWS)):
				if GUI._VIEWS[ii].lower() == vw:
					cls._view_idx = ii
					break
		cls._eventbox = Gtk.EventBox()
		cls._box_view_cont.pack_start(cls._eventbox, True, True, 0)
		cls._box_view_cont.reorder_child(cls._eventbox, view_pos)
		cls._eventbox.add(cls.view_widgets[cls._view_idx])
		cls._eventbox.connect('key_press_event', cls.keypress)
		cls.view_widgets[cls._view_idx].grab_focus() # so it gets keypresses
		del(cls._box_view_cont) # don't need this anymore

		# Add functionality to spinbuttons not provided by GTK
		cls._init_spinbuttons()
		cls._init_comboboxes()
		cls._init_entryboxes()

		# Menu bar & softbutton bar made insensitive in .glade for startup.
		# We make them sensitive here before activating view.
		cls._builder.get_object('menu_bar').set_sensitive(True)
		cls._builder.get_object('box_buttons').set_sensitive(True)

		cls.view_redraw() # Draw active view
		cls._eventbox.show_all()


	@classmethod
	def _init_locale(cls) -> None:
		# Initialise language from config or OS settings
		lang = Config.get('global', 'language')
		if not lang:
			lang = locale.getlocale()
		if isinstance(lang,str) and '.' not in lang:
			# Need to include encoding
			lang = (lang,'UTF-8')
		locale.setlocale(locale.LC_ALL, lang)
		locale.bindtextdomain('pygenda', cls._LOCALE_DIR)
		locale.textdomain('pygenda')


	@classmethod
	def _init_clipboard(cls) -> None:
		# Load clipboard helper library if available, or set to None
		try:
			libclip_file = '{:s}/libpygenda_clipboard.so'.format(ospath.dirname(__file__))
			cls.lib_clip = ctypes.CDLL(libclip_file)
		except:
			print('Warning: Failed to load clipboard library', file=stderr)
			cls.lib_clip = None


	@classmethod
	def _init_date_format(cls) -> None:
		# Initialise date formatting strings from config
		cls.date_order = cls._date_order_from_config()

		# Make date_formatting_numeric - a format string like '%Y-$m-$d'
		dto_tmp = cls.date_order.replace('M','m').replace('D','d')
		cls.date_formatting_numeric = '%{0:s}{sep:s}%{1:s}{sep:s}%{2:s}'.format(dto_tmp[0],dto_tmp[1],dto_tmp[2],sep=Config.get('global','date_sep'))

		# Make format strings for longer formats, e.g. "Mon Dec 31, 2001"
		cls.date_formatting_text = Config.get('global','date_fmt_text')

		# Other format strings might be constructed if not set in config file
		# date_formatting_text_noyear
		cst = Config.get('global','date_fmt_text_noyear')
		cls.date_formatting_text_noyear = ' '.join(cls.date_formatting_text.replace('%y','%Y').replace('%Y,',' ').replace(', %Y',' ').replace('%Y','').split()) if not cst else cst

		# date_formatting_textabb
		cst = Config.get('global','date_fmt_textabb')
		cls.date_formatting_textabb = cls.date_formatting_text.replace('%A','%a').replace('%B','%b') if not cst else cst

		# date_formatting_textabb_noyear
		cst = Config.get('global','date_fmt_textabb_noyear')
		cls.date_formatting_textabb_noyear = ' '.join(cls.date_formatting_textabb.replace('%y','%Y').replace('%Y,',' ').replace(', %Y',' ').replace('%Y','').split()) if not cst else cst


	@staticmethod
	def _date_order_from_config() -> str:
		# Process date order string from config.
		# Converts, e.g. 'YYYY-MM-DD' -> 'YMD' & checks output is valid.
		raw = Config.get('global','date_ord').upper()
		ret = ''
		for ch in raw:
			if ch in 'YMD' and ch not in ret:
				ret += ch
		assert(len(ret)==3)
		return ret


	@classmethod
	def _init_views(cls) -> None:
		# Get new Gtk Widgets for views.
		# Add view switching options to menu.
		cls.views = []
		cls.view_widgets = []
		for v in GUI._VIEWS:
			m = import_module('.pygenda_view_{:s}'.format(v.lower()),package='pygenda')
			cls.views.append(getattr(m, 'View_{:s}'.format(v)))
			cls.view_widgets.append(cls.views[-1].init())
			cls.view_widgets[-1].get_style_context().add_class('view')

		menu_views_list = cls._builder.get_object('menu_views_list')
		accel_gp = Gtk.AccelGroup()
		cls._window.add_accel_group(accel_gp)
		for i in range(len(cls.views)):
			m = Gtk.MenuItem(cls.views[i].view_name())
			m.set_use_underline(True)
			akey = cls.views[i].accel_key()
			if akey:
				m.add_accelerator('activate', accel_gp, akey, Gdk.ModifierType.SHIFT_MASK|Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)
			m.connect('activate',cls.switch_view,i)
			m.show()
			menu_views_list.add(m)


	@classmethod
	def _init_spinbuttons(cls) -> None:
		# Connect spinbutton keypress event to handler for extra features.
		# All spinbuttons are part of the Entry dialog, so this could go
		# in the EntryDialogController class. However, in theory spinbuttons
		# can be anywhere, so putting this in the general GUI class.
		for sb_id in ('allday_count','repeat_interval','repeat_occurrences'):
			sb = cls._builder.get_object(sb_id)
			sb.connect('key-press-event', cls._spinbutton_keypress)
			sb.connect('focus-out-event', cls._focusout_unhighlight)


	@classmethod
	def _init_comboboxes(cls) -> None:
		# Connect ComboBox events to handlers for extra features.
		for cb_id in ('combo_repeat_type','combo_bydaymonth','combo_byday_ord','combo_byday_day','combo_status'):
			cb = cls._builder.get_object(cb_id)
			cb.connect('key-press-event', cls._combobox_keypress)


	@classmethod
	def _init_entryboxes(cls) -> None:
		# Connect Entry textbox events to handlers for extra features.
		for eb_id in ('entry_desc','entry_location'):
			eb = cls._builder.get_object(eb_id)
			eb.connect('focus-out-event', cls._focusout_unhighlight)


	@classmethod
	def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> None:
		# Called whenever a key is pressed/repeated when View in focus
		cls.views[cls._view_idx].keypress(wid,ev)


	@staticmethod
	def _spinbutton_keypress(wid:Gtk.SpinButton, ev:Gdk.EventKey) -> bool:
		# Called to handle extra spinbutton keyboard controls
		shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
		if ev.keyval in GUI.SPINBUTTON_INC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Up):
			wid.update() # So if user types "5" then "+" value changes to "6"
			wid.spin(Gtk.SpinType.STEP_FORWARD, 1)
			return True # done
		if ev.keyval in GUI.SPINBUTTON_DEC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Down):
			wid.update()
			wid.spin(Gtk.SpinType.STEP_BACKWARD, 1)
			return True # done
		if ev.keyval==Gdk.KEY_Up:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.UP)
		if ev.keyval==Gdk.KEY_Down:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.DOWN)
		return False # propagate event


	@staticmethod
	def _combobox_keypress(wid:Gtk.ComboBox, ev:Gdk.EventKey) -> bool:
		# Called to handle extra combobox keyboard controls
		# BUG!! This is not called when the combobox is in "popout" state
		if ev.keyval==Gdk.KEY_Return:
			# Manually trigger default event on dialog box
			dlg = wid.get_toplevel()
			if dlg:
				dlg.response(Gtk.ResponseType.OK)
			return True # done

		mdl = wid.get_model()
		count = mdl.iter_n_children()
		a = wid.get_active()
		shiftdown = ev.state&Gdk.ModifierType.SHIFT_MASK
		if ev.keyval in GUI.SPINBUTTON_INC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Down):
			a = (a+1)%count
			wid.set_active(a)
			return True # done
		if ev.keyval in GUI.SPINBUTTON_DEC_KEY or (shiftdown and ev.keyval==Gdk.KEY_Up):
			if a<0:
				a = 0
			a = (a-1)%count
			wid.set_active(a)
			return True # done

		if ev.keyval==Gdk.KEY_Up:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.UP)
		if ev.keyval==Gdk.KEY_Down:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.DOWN)
		if ev.keyval in (Gdk.KEY_Left,Gdk.KEY_Right):
			# Return - otherwise detected as alphabetic (not sure why!)
			return False

		ch = chr(ev.keyval)
		if ch.isalpha() or ch.isdigit():
			ch = ch.lower()
			# Search for ch in first characters of combobox values
			it = mdl.iter_nth_child(None,a)
			for i in range(1,count):
				it = mdl.iter_next(it)
				if it is None: # at end of list, loop to top!
					it = mdl.get_iter_first()
				ch1 = mdl.get(it,0)[0][0].lower()
				if ch==ch1:
					wid.set_active((a+i)%count)
					break
			return True # done

		return False # propagate event


	@staticmethod
	def _focusout_unhighlight(wid:Gtk.Widget, ev:Gdk.EventKey) -> bool:
		# Called to handle remove highlight from entry box/spinbutton
		# when focus moves to another widget
		dlg = wid.get_toplevel()
		if dlg.get_focus() != wid:
			wid.select_region(0,0) # remove highlight
		return False # propagate event


	@classmethod
	def view_redraw(cls, ev_changes:bool=False) -> None:
		# Redraw the currently active view.
		# ev_changes: bool, True if events need updating too
		cls.views[cls._view_idx].redraw(ev_changes)


	@classmethod
	def switch_view(cls, wid:Gtk.Widget, idx:int=None) -> None:
		# Callback from UI widget (e.g. menu, softbutton) to change view.
		# idx = index of new view (otherwise goes to next view in list)
		if idx is None:
			# Go to next view in list
			cls._view_idx = (cls._view_idx+1)%len(cls.views)
		elif cls._view_idx == idx:
			return # No change, so skip redraw
		else:
			cls._view_idx = idx
		cls._eventbox.remove(cls._eventbox.get_child())
		new_view = cls.views[cls._view_idx]
		new_view.renew_display()
		new_view.redraw(True)
		new_wid = cls.view_widgets[cls._view_idx]
		cls._eventbox.add(new_wid)
		new_wid.grab_focus()
		new_wid.show_all()


	@classmethod
	def cursor_set(cls, dt:dt_date, idx:int=None) -> None:
		# Set current cursor date, and optionally the index within the date.
		# Call redraw on view if required.
		if dt != cls.cursor_date or (idx is not None and idx != cls.cursor_idx_in_date):
			cls.cursor_date = dt
			if idx is not None:
				cls.cursor_idx_in_date = idx
			cls.view_redraw(False)


	@classmethod
	def cursor_inc(cls, delta:timedelta, idx:int=None) -> None:
		# Add delta to current cursor date; optionally set index in date.
		# Call redraw on view.
		cls.cursor_date += delta
		if idx is not None:
			cls.cursor_idx_in_date = idx
		cls.view_redraw(False)


	# Main
	@classmethod
	def main(cls) -> None:
		# Run the man Gtk loop
		Gtk.main()


	# Signal handling functions
	@classmethod
	def exit(cls, *args) -> None:
		# Callback for various types of exit signal (command line, menus...)
		Gtk.main_quit()

	@classmethod
	def event_newentry(cls, *args) -> None:
		# Callback for new entry signal (menu, softbutton)
		EntryDialogController.newentry()

	@classmethod
	def event_edittime(cls, *args) -> None:
		# Callback for change-entry-time signal
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en:
			EntryDialogController.editentry(en, EntryDialogController.TAB_TIME)

	@classmethod
	def event_editrepeats(cls, *args) -> None:
		# Callback for change-entry-repeats signal
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en:
			EntryDialogController.editentry(en, EntryDialogController.TAB_REPEATS)

	@classmethod
	def event_editalarm(cls, *args) -> None:
		# Callback for change-entry-alarm signal
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en:
			EntryDialogController.editentry(en, EntryDialogController.TAB_ALARM)

	@classmethod
	def event_editdetails(cls, *args) -> None:
		# Callback for change-entry-details signal
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en:
			EntryDialogController.editentry(en, EntryDialogController.TAB_DETAILS)


	@classmethod
	def event_stat_toggle(cls, stat:Optional[str]) -> None:
		# Handle signals from menu to change current entry's status.
		# Paramenter stat is None or text string for status (eg 'CONFIRMED').
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en:
			Calendar.set_toggle_status_entry(en, stat)
			cls.view_redraw(True)


	@classmethod
	def delete_request(cls, *args) -> None:
		# Callback to implement "delete" from GUI, e.g. backspace key pressed
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en is not None:
			cls.dialog_deleteentry(en)


	@classmethod
	def cut_request(cls, *args) -> None:
		# Handler to implement "cut" from GUI, e.g. cut clicked in menu
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en and 'SUMMARY' in en:
			if cls.lib_clip is None:
				# Don't do fallback - might lead to unexpected data loss
				print('Warning: No clipboard library, cut not available', file=stderr)
			elif 'RRULE' in en: # repeating entry
				# Need to think about how to implement this from UI side.
				# Problem: Does user expect single occurrence to be cut, or all?
				# Maybe bring up dialog "Cut single occurrence, or all repeats?"
				# Then, do we do the came for Copying repeating entries?
				# How do we adapt repeats when moved to a different date?
				print('Warning: Cutting repeating entries not implemented', file=stderr)
			else:
				txtbuf = bytes(en['SUMMARY'], 'utf-8')
				calbuf = en.to_ical()
				cls.lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))
				Calendar.delete_entry(en)
				cls.view_redraw(True)


	@classmethod
	def copy_request(cls, *args) -> None:
		# Handler to implement "copy" from GUI, e.g. copy clicked in menu
		en = cls.views[cls._view_idx].get_cursor_entry()
		if en and 'SUMMARY' in en:
			if cls.lib_clip is None:
				print('Warning: No clipboard library, fallback to text copy', file=stderr)
				cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
				txt = en['SUMMARY']
				cb.set_text(txt, -1)
			else:
				txtbuf = bytes(en['SUMMARY'], 'utf-8')
				calbuf = en.to_ical()
				cls.lib_clip.set_cb(ctypes.create_string_buffer(txtbuf),ctypes.create_string_buffer(calbuf))


	@classmethod
	def paste_request(cls, *args) -> None:
		# Handler to implement "paste" from GUI, e.g. paste clicked in menu
		cb = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

		# First, we try requesting a 'text/calendar' type from the clipboard
		sdat = cb.wait_for_contents(Gdk.Atom.intern('text/calendar', False))
		try:
			events = iEvent.from_ical(sdat.get_data())
			ev = events.walk('VEVENT')[0]
			Calendar.new_entry_from_example(ev, dt_start=cls.cursor_date)
			cls.view_redraw(True)
			return
		except:
			None

		# Fallback: request plain text from clipboard
		txt = cb.wait_for_text()
		if txt is not None:
			txt = txt.strip()
			txt = txt.replace('\n',' ')
			txt = txt.replace('\t',' ')
			# Open a New Entry dialog with description initialised as txt
			GLib.idle_add(EntryDialogController.newentry,txt)


	@classmethod
	def dialog_deleteentry(cls, en:iEvent) -> None:
		# Dialog to implement "delete" from GUI, e.g. backspace key
		dialog = Gtk.Dialog(title=_('Delete Entry'), parent=cls._window,
		    flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
		    buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE, Gtk.STOCK_DELETE, Gtk.ResponseType.APPLY))
		if 'RRULE' in en:
			# repeating entry - clarify what is being deleted
			# !! We should really ask if user wants to delete all/single etc.
			l_template = _('Delete all repeats:\n"{:s}"?')
		else:
			l_template = _('Delete entry:\n"{:s}"?')
		lab = Gtk.Label(l_template.format(en['SUMMARY'] if 'SUMMARY' in en else 'UNNAMED'))
		if (not dialog or not lab): # Sanity check
			raise NameError('Dialog Delete creation failure')
		dialog.set_resizable(False)
		lab.set_justify(Gtk.Justification.CENTER)
		dialog.get_content_area().add(lab)
		dialog.set_default_response(Gtk.ResponseType.APPLY)#Enter action
		dialog.show_all()
		response = dialog.run()
		dialog.destroy()
		if response == Gtk.ResponseType.APPLY:
			Calendar.delete_entry(en)
			cls.view_redraw(True)


	@classmethod
	def dialog_goto(cls, *args) -> None:
		# Called to implement "go to" from GUI, e.g. button
		dialog = Gtk.Dialog(title=_('Go To'), parent=cls._window,
		    flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
		    buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE))
		wdate = WidgetDate(cls.cursor_date)
		if (not dialog or not wdate): # Allocation check
			raise NameError('Dialog Goto creation failure')
		wdate.connect('changed', GUI.check_date_fixed)
		dialog.set_resizable(False)
		dialog.get_content_area().add(wdate)
		wdate.set_halign(Gtk.Align.CENTER)
		icon = Gtk.Image()
		icon.set_from_stock(Gtk.STOCK_JUMP_TO, Gtk.IconSize.BUTTON)
		go_but = Gtk.Button(label=_('Go'), image=icon)
		go_but.set_can_default(True)
		dialog.add_action_widget(go_but, Gtk.ResponseType.APPLY)
		go_but.grab_default() # So "Enter" activates go_but
		dialog.show_all()
		while True:
			response = dialog.run()
			dt = wdate.get_approx_date_or_none() # So Feb 30th -> end of Feb
			if response!=Gtk.ResponseType.APPLY or dt is not None:
				break
			# Date is invalid, add error styling
			wdate.get_style_context().add_class(GUI.STYLE_ERR)
		if response == Gtk.ResponseType.APPLY:
			cls.cursor_set(dt)
		dialog.destroy()


	@staticmethod
	def check_date_fixed(wid:WidgetDate) -> None:
		# Removes error highlight if date is valid (e.g. not 30th Feb)
		# Would be nice to make this a method of WidgetDate class.
		# Can be used as a callback, e.g. attached to 'changed' signal
		if wid.get_date_or_none() is not None:
			wid.get_style_context().remove_class(GUI.STYLE_ERR)


	@classmethod
	def dialog_about(cls, *args) -> None:
		# Display the "About" dialog
		dialog = Gtk.AboutDialog(parent=cls._window)
		dialog.set_program_name('Pygenda')
		dialog.set_copyright(u'Copyright © 2022 Matthew Lewis')
		dialog.set_license_type(Gtk.License.GPL_3_0_ONLY)
		dialog.set_logo_icon_name('x-office-calendar')
		dialog.set_authors(('Matthew Lewis',))
		dialog.set_version('version {:s}'.format(__version__))
		dialog.set_comments(_(u'A calendar/agenda application written in Python/GTK3. The UI is inspired by the Agenda programs on the Psion Series 3 and Series 5 PDAs.\nWARNING: This is in-development code, released as a preview for developers. It will probably corrupt your data.'))
		dialog.show_all()
		dialog.run()
		dialog.destroy()


	@classmethod
	def toggle_fullscreen(cls, *args) -> None:
		# Callback to toggle fullscreen mode on/off (e.g. from menu)
		# !! to-do: update menu text to reflect state
		cls._is_fullscreen = not cls._is_fullscreen
		if cls._is_fullscreen:
			cls._window.fullscreen()
		else:
			cls._window.unfullscreen()


	@classmethod
	def debug(cls, *args):
		# Temporary callback - delete me !!!!
		# Placeholder until we decide what fourth button does
		print('Button clicked {}'.format(args[0]))


# Singleton class to manage Entry dialog
class EntryDialogController:
	TAB_TIME = 0
	TAB_REPEATS = 1
	TAB_ALARM = 2
	TAB_DETAILS = 3
	TAB_COUNT = 4

	@classmethod
	def init(cls):
		# Initialiser for singleton class.
		# Called from GUI init_stage2().

		# Get some references to dialog elements in glade
		cls.dialog = GUI._builder.get_object('dialog_entry')
		if (not cls.dialog): # Sanity check
			raise NameError('Dialog Entry not found')

		cls.wid_desc = GUI._builder.get_object('entry_desc')
		cls._wid_desc_changed_handler = cls.wid_desc.connect('changed', cls._desc_changed)
		wid_grid = GUI._builder.get_object('dialogentry_grid')
		cls.wid_tabs = GUI._builder.get_object('entry_tabs')

		# Create & add date entry widget
		cls.wid_date = WidgetDate(GUI.cursor_date)
		wid_grid.attach(cls.wid_date, 1,1, 1,1)# Can we locate this another way?
		cls.wid_date.show_all()

		# Init other groups of widgets
		cls._init_timefields()
		cls._init_repeatfields()
		cls._init_alarmfields()
		cls._init_detailfields()
		# Also need to init alarm & detail fields
		cls._init_navigation()

		# cls._empty_desc_allowed has three possible values:
		#   None - empty desc allowed in dialog (signals delete entry)
		#   True - empty desc allowed, but can turn False!
		#   False - empty desc not allowed in dialog (e.g. new entry)
		cls._empty_desc_allowed = None


	@classmethod
	def _init_timefields(cls) -> None:
		# Initialise widgets etc in the Entry dialog under the "Time" tab.
		# Called on app startup.
		cls.wid_timed_buttons = GUI._builder.get_object('radiobuttons_istimed').get_children()

		# Create widgets for entry time & duration fields
		cls.wid_time = WidgetTime(dt_time(hour=9))
		cls.wid_dur = WidgetDuration(timedelta())
		cls.wid_endtime = WidgetTime(dt_time(hour=9))
		# Add to dialog
		wid_reveal_tm_e = GUI._builder.get_object('revealer_time_e')
		tbox = wid_reveal_tm_e.get_child() # should be a GtkBox
		tbox.add(cls.wid_time)
		tbox.add(cls.wid_dur)
		tbox.add(cls.wid_endtime)
		tbox.show_all()

		# Make lists of revealers, so they can be hidden/revealed
		cls.revs_timedur = cls._revealers_from_ids('revealer_time_l', 'revealer_time_e')
		cls.revs_allday = cls._revealers_from_ids('revealer_allday_l', 'revealer_allday_e')
		cls.wid_allday_count = GUI._builder.get_object('allday_count')

		cls.wid_timed_buttons[1].connect('toggled', cls._do_timed_toggle)
		cls.wid_timed_buttons[2].connect('toggled', cls._do_allday_toggle)

		# We keep references to these connections, so then can be (un)blocked.
		cls._tmdur_handler_time = cls.wid_time.connect('changed', cls._tmdur_changed, 0)
		cls._tmdur_handler_dur = cls.wid_dur.connect('changed', cls._tmdur_changed, 1)
		cls._tmdur_handler_endtime = cls.wid_endtime.connect('changed', cls._tmdur_changed, 2)

		# By default the signal is *blocked* - unblock when dialog active
		cls._block_tmdur_signals()


	@classmethod
	def _init_repeatfields(cls) -> None:
		# Initialise widgets etc in the Entry dialog under the "Repeats" tab.
		# Called on app startup.
		cls.wid_rep_type = GUI._builder.get_object('combo_repeat_type')
		cls.revs_repeat = cls._revealers_from_ids('revealer_repeat_l','revealer_repeat_e')
		cls.revs_rep_monthdays = cls._revealers_from_ids('revealer_repeat_monthday_e')
		cls.revs_rep_weekdays = cls._revealers_from_ids('revealer_repeat_weekday_e')
		cls.revs_rep_monthweekdays = cls._revealers_from_ids('revealer_repeat_day_l')
		cls.wid_rep_type.connect('changed', cls._reptype_changed)

		cls.wid_rep_interval = GUI._builder.get_object('repeat_interval')
		cls.wid_rep_forever = GUI._builder.get_object('repeat_forever')
		cls.wid_repbymonthday = GUI._builder.get_object('combo_bydaymonth')
		cls.wid_repbyweekday_day = GUI._builder.get_object('combo_byday_day')
		cls.wid_repbyweekday_ord = GUI._builder.get_object('combo_byday_ord')
		cls.wid_rep_forever.connect('toggled', cls._do_repeatforever_toggle)

		cls.wid_rep_occs = GUI._builder.get_object('repeat_occurrences')
		cls.wid_rep_enddt = WidgetDate(GUI.cursor_date)
		rbox = GUI._builder.get_object('revealer_repeat_until_e').get_child() # Should be a GtkBox
		rbox.add(cls.wid_rep_enddt)
		rbox.show_all()
		cls.revs_rep_ends = cls._revealers_from_ids('revealer_repeat_until_l', 'revealer_repeat_until_e')

		cls.rep_occs_determines_end = True
		cls._rep_handler_date = cls.wid_date.connect('changed', cls._repend_changed,0)
		cls._rep_handler_time = cls.wid_time.connect('changed', cls._repend_changed,0)
		cls._rep_handler_type = cls.wid_rep_type.connect('changed', cls._repend_changed,0)
		cls._rep_handler_mday = cls.wid_repbymonthday.connect('changed', cls._repend_changed,0)
		cls._rep_handler_wdayday = cls.wid_repbyweekday_day.connect('changed', cls._repend_changed,0)
		cls._rep_handler_wdayord = cls.wid_repbyweekday_ord.connect('changed', cls._repend_changed,0)
		cls._rep_handler_inter = cls.wid_rep_interval.connect('changed', cls._repend_changed,0)
		cls._rep_handler_occs = cls.wid_rep_occs.connect('changed', cls._repend_changed,1)
		cls._rep_handler_enddt = cls.wid_rep_enddt.connect('changed', cls._repend_changed,2)

		# For repeat on "1st Sat." etc, we fill ComboBox with local day names
		day = Config.get_int('global','start_week_day')
		for i in range(7):
			cls.wid_repbyweekday_day.append(RepeatInfo.DAY_ABBR[day],day_name[day])
			day = (day+1)%7

		# Get label used to display exception dates
		cls._lab_rep_exceptions = GUI._builder.get_object('lab_rep_exceptions')

		# We want default to be signal *blocked* - unblock when dialog active
		cls._block_rep_occend_signals()


	@classmethod
	def _init_alarmfields(cls) -> None:
		# Initialise widgets etc in the Entry dialog under the "Alarm" tab.
		# Called on app startup.
		cls.wid_alarmset = GUI._builder.get_object('alarm-set')


	@classmethod
	def _init_detailfields(cls) -> None:
		# Initialise widgets etc in the Entry dialog under the "Details" tab.
		# Called on app startup.
		cls.wid_status = GUI._builder.get_object('combo_status')
		cls.wid_location = GUI._builder.get_object('entry_location')


	@classmethod
	def _init_navigation(cls) -> None:
		# For some reason, cannot generally navigate up/down from
		# radio buttons. This sets up a custom handler.
		for i in range(3):
			cls.wid_timed_buttons[i].connect('key_press_event', cls._radiobutton_keypress)


	@classmethod
	def _revealers_from_ids(cls, *idlist) -> list:
		# Given a list of id strings, in glade file, make a list of revealers
		revs = []
		for id in idlist:
			obj = GUI._builder.get_object(id)
			if obj is not None:
				revs.append(obj)
				# Reduce transition time, o/w devices like Gemini feel sluggish.
				# Tried to do this with CSS, so could be set per device,
				# but couldn't get it to work. So this is a workaround.
				# Alternative: in ~/.config/gtk-3.0/settings.ini, try one of
				#   gtk-enable-animations=0
				#   gtk-revealer-transition-duration=10
				obj.set_transition_duration(20) # millisecs; default 250
			else:
				print('Warning: Object with id "{:s}" not found'.format(id), file=stderr)
		return revs


	@staticmethod
	def _do_multireveal(revlist, reveal:bool) -> None:
		# Given a list of ids, reveal (or hide) children
		for r in revlist:
			r.set_reveal_child(reveal)


	@staticmethod
	def _radiobutton_keypress(wid:Gtk.RadioButton, ev:Gdk.EventKey) -> bool:
		# Custom hander for up/down keys on radiobuttons.
		# Note: Assumes there's nothing to left/right so can use TAB_FWD/BKWD
		if ev.keyval==Gdk.KEY_Up:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.TAB_BACKWARD)
		if ev.keyval==Gdk.KEY_Down:
			return wid.get_toplevel().child_focus(Gtk.DirectionType.TAB_FORWARD)
		return False # Indicates event still needs handling - propagate it


	@classmethod
	def _cancel_empty_desc_allowed(cls) -> None:
		# Change _empty_desc_allowed True->False
		# Leave unchanged if is None
		if cls._empty_desc_allowed:
			cls._empty_desc_allowed = False


	@classmethod
	def _do_timed_toggle(cls, wid:Gtk.RadioButton) -> None:
		# Callback. Called when "timed" radio button changes state.
		# Reveals/hides appropriate sub-obtions, and flags that
		# dialog state had been changed.
		ti = cls.wid_timed_buttons[1].get_active()
		cls._do_multireveal(cls.revs_timedur, ti)
		cls._cancel_empty_desc_allowed()


	@classmethod
	def _do_allday_toggle(cls, wid:Gtk.RadioButton) -> None:
		# Callback. Called when "all day" radio button changes state.
		# Reveals/hides appropriate sub-obtions, and flags that
		# dialog state had been changed.
		ad = cls.wid_timed_buttons[2].get_active()
		cls._do_multireveal(cls.revs_allday, ad)
		cls._cancel_empty_desc_allowed()


	@classmethod
	def _desc_changed(cls, wid:Gtk.Entry) -> None:
		# Callback. Called when entry description is changed by user.
		# Flags that dialog state had been changed.
		# Removes error state styling if it is no longer needed.
		if cls.wid_desc.get_text(): # if desc field is non-empty
			cls._cancel_empty_desc_allowed()
		cls._is_valid_entry(set_style=False) # remove error styles if present


	@classmethod
	def _reptype_changed(cls, wid:Gtk.ComboBox) -> None:
		# Callback. Called when entry repeat type is changed by user.
		# Reveals/hides relevant sub-options.
		# wid should be the repeat-type combobox
		st = wid.get_active()>0
		cls._do_multireveal(cls.revs_repeat, st)
		monthday = (wid.get_active_id()=='MONTHLY-MONTHDAY')
		weekday = (wid.get_active_id()=='MONTHLY-WEEKDAY') # Booleans
		cls._do_multireveal(cls.revs_rep_monthdays, monthday)
		cls._do_multireveal(cls.revs_rep_weekdays, weekday)
		cls._do_multireveal(cls.revs_rep_monthweekdays, monthday or weekday)
		if monthday and not cls.repbymonthday_initialized:
			# First time shown showing monthday-repeats, so initialise based on start date
			cls.repbymonthday_initialized = True
			sdt = cls.get_date_start()
			if sdt is None: # Fallback in case date is invalid
				cls.wid_repbymonthday.set_active(0)
			else:
				idx = sdt.day - monthrange(sdt.year,sdt.month)[1] - 1
				if -7 <= idx <= -2:
					cls.wid_repbymonthday.set_active_id(str(idx))
				else:
					cls.wid_repbymonthday.set_active(0)
		elif weekday and not cls.repbyweekday_initialized:
			# First time shown showing weekday-repeats, so initialise based on start date
			cls.repbyweekday_initialized = True
			sdt = cls.get_date_start()
			if sdt is None: # Fallback in case date is invalid
				cls.wid_repbyweekday_ord.set_active(0)
				cls.wid_repbyweekday_day.set_active(0)
			else:
				wkst = Config.get_int('global','start_week_day')
				cls.wid_repbyweekday_day.set_active((sdt.weekday()-wkst)%7)
				if sdt.day<=21:
					# Value will be, e.g. "1st", "2nd" (Tues of month)
					cls.wid_repbyweekday_ord.set_active_id(str(1+(sdt.day-1)//7))
				else:
					# Want, e.g. "last" (Friday of month)
					rem = monthrange(sdt.year,sdt.month)[1]-sdt.day
					cls.wid_repbyweekday_ord.set_active_id(str(-1-(rem//7)))
		cls._cancel_empty_desc_allowed()


	@classmethod
	def _do_repeatforever_toggle(cls, wid:Gtk.Button) -> None:
		# Callback. Called when repeat-forever state is changed by user.
		# Reveals/hides relevant sub-options.
		st = not wid.get_active() #If *not* forever, we show repeat-til options
		cls._do_multireveal(cls.revs_rep_ends, st)
		if st:
			# Want to make sure revealed elements are synced
			if cls.rep_occs_determines_end:
				cls._sync_rep_occs_end()
			else:
				# if enddate determines occs, need to check enddate>=startdate
				sdt = cls.get_date_start()
				edt = cls.wid_rep_enddt.get_date_or_none()
				if sdt is not None and (edt is None or edt < sdt):
					cls.wid_rep_enddt.set_date(sdt)
					cls.wid_rep_occs.set_value(1)

		cls._cancel_empty_desc_allowed()


	@classmethod
	def _block_tmdur_signals(cls) -> None:
		# Block update signals from start time, duration & endtime
		# fields, so they can be updated programmatically without
		# getting "update signal feedback loops".
		cls.wid_time.handler_block(cls._tmdur_handler_time)
		cls.wid_dur.handler_block(cls._tmdur_handler_dur)
		cls.wid_endtime.handler_block(cls._tmdur_handler_endtime)


	@classmethod
	def _unblock_tmdur_signals(cls) -> None:
		# Unblock update signals from start time, duration & endtime
		# fields, so if user changes one then other fields update.
		cls.wid_time.handler_unblock(cls._tmdur_handler_time)
		cls.wid_dur.handler_unblock(cls._tmdur_handler_dur)
		cls.wid_endtime.handler_unblock(cls._tmdur_handler_endtime)


	@classmethod
	def _block_rep_occend_signals(cls) -> None:
		# Function to disable "changed" signals from various widgets that
		# lead to either the end-date or repeat count being re-calculated
		# based on the new values. Generally, keep signals blocked except
		# when the dialog is being shown to the user.
		cls.wid_date.handler_block(cls._rep_handler_date)
		cls.wid_time.handler_block(cls._rep_handler_time)
		cls.wid_rep_type.handler_block(cls._rep_handler_type)
		cls.wid_repbymonthday.handler_block(cls._rep_handler_mday)
		cls.wid_repbyweekday_day.handler_block(cls._rep_handler_wdayday)
		cls.wid_repbyweekday_ord.handler_block(cls._rep_handler_wdayord)
		cls.wid_rep_interval.handler_block(cls._rep_handler_inter)
		cls.wid_rep_occs.handler_block(cls._rep_handler_occs)
		cls.wid_rep_enddt.handler_block(cls._rep_handler_enddt)


	@classmethod
	def _unblock_rep_occend_signals(cls) -> None:
		# Function to re-enable "changed" signals from various widgets that
		# lead to either the end-date or repeat count being re-calculated
		# based on the new values. Generally, unblock when the dialog is
		# being used and user can change these values.
		cls.wid_date.handler_unblock(cls._rep_handler_date)
		cls.wid_time.handler_unblock(cls._rep_handler_time)
		cls.wid_rep_type.handler_unblock(cls._rep_handler_type)
		cls.wid_repbymonthday.handler_unblock(cls._rep_handler_mday)
		cls.wid_repbyweekday_day.handler_unblock(cls._rep_handler_wdayday)
		cls.wid_repbyweekday_ord.handler_unblock(cls._rep_handler_wdayord)
		cls.wid_rep_interval.handler_unblock(cls._rep_handler_inter)
		cls.wid_rep_occs.handler_unblock(cls._rep_handler_occs)
		cls.wid_rep_enddt.handler_unblock(cls._rep_handler_enddt)


	@classmethod
	def _tmdur_changed(cls, wid:Gtk.Widget, el:int) -> None:
		# Callback. Called when starttime/endtime/duration changed.
		# 'el' paramenter indicates which one changed (but could use wid).
		sdt = cls.get_date_start()
		stm = cls.wid_time.get_time_or_none()
		d = cls.wid_dur.get_duration_or_none()
		etm = cls.wid_endtime.get_time_or_none()
		if sdt is not None and stm is not None and d is not None and etm is not None:
			# First, update which of dur & end-time determines the other
			if el==1:
				cls.dur_determines_end = True
			elif el==2:
				cls.dur_determines_end = False

			# Second, do the required update
			cls._block_tmdur_signals() # prevent accidental recursion
			try:
				st_dttm = dt_datetime.combine(sdt,stm)
				if cls.dur_determines_end:
					cls.wid_endtime.set_time(st_dttm+d)
				else: # need to update dur
					end_dttm = dt_datetime.combine(sdt,etm)
					if end_dttm<st_dttm:
						end_dttm += timedelta(days=1)
					cls.wid_dur.set_duration(end_dttm - st_dttm)
			finally:
				cls._unblock_tmdur_signals()
		cls._cancel_empty_desc_allowed()
		cls._is_valid_entry(set_style=False) # remove error styles if present


	@classmethod
	def _repend_changed(cls, wid:Gtk.Widget, el:int) -> None:
		# Handler for signals from any widget that might result in either the
		# repeat end-date or occurence count changing. Prompts recalculation
		# and possibly also updates which of end-date/occurences is the leader.
		# Variable el indicates if either possible leader was updated.
		if el == 1: # Occurrences edited, make it "leader"
			cls.rep_occs_determines_end = True
			cls._set_occs_min(1)
		elif el == 2: # Repeat until edited, make it "leader"
			cls.rep_occs_determines_end = False

		cls._block_rep_occend_signals() # prevent accidental recursion
		try:
			cls._sync_rep_occs_end()
		finally:
			cls._unblock_rep_occend_signals()
		cls._cancel_empty_desc_allowed()
		cls._is_valid_entry(set_style=False) # remove error styles if present


	@classmethod
	def _set_occs_min(cls, mn:int) -> None:
		# quick method to set minimum of the occurrences spinbutton
		mx = cls.wid_rep_occs.get_range().max
		cls.wid_rep_occs.set_range(mn,mx)


	# Maps used to calculate repeats using dateutil.rrule class
	MAP_RTYPE_TO_RRULE = {
		'YEARLY': du_rrule.YEARLY,
		'MONTHLY': du_rrule.MONTHLY,
		'MONTHLY-MONTHDAY': du_rrule.MONTHLY,
		'MONTHLY-WEEKDAY': du_rrule.MONTHLY,
		'WEEKLY': du_rrule.WEEKLY,
		'DAILY': du_rrule.DAILY,
		'HOURLY': du_rrule.HOURLY,
		'MINUTELY': du_rrule.MINUTELY,
		'SECONDLY': du_rrule.SECONDLY
		}

	MAP_RDAY_TO_RRULEDAY = {
		'MO': du_rrule.MO,
		'TU': du_rrule.TU,
		'WE': du_rrule.WE,
		'TH': du_rrule.TH,
		'FR': du_rrule.FR,
		'SA': du_rrule.SA,
		'SU': du_rrule.SU
		}

	@classmethod
	def _sync_rep_occs_end(cls) -> None:
		# Function to recalculate repeat end-date or occurences count.
		# Called when widgets that might affect repeat counts are updated.
		if cls.wid_rep_forever.get_active():
			return
		rtype = cls.wid_rep_type.get_active_id()
		if rtype is None:
			return
		if cls.rep_occs_determines_end:
			stdt = cls.get_date_start()
			if stdt is not None:
				span = (cls.get_repeat_occurrences()-1) * cls.get_repeat_interval()
				if rtype=='YEARLY':
					if stdt.day==29 and stdt.month==2: # 29th Feb - leap day!
						cls._sync_rep_end_from_occ_rrule(rtype)
						return
					delta = relativedelta(years=span)
				elif rtype=='MONTHLY':
					if cls.get_datetime_start().day>=29:
						cls._sync_rep_end_from_occ_rrule(rtype)
						return
					else:
						delta = relativedelta(months=span)
				elif rtype=='WEEKLY':
					delta = timedelta(days=span*7)
				elif rtype=='DAILY':
					delta = timedelta(days=span)
				elif rtype=='HOURLY':
					delta = timedelta(hours=span)
				elif rtype=='MINUTELY':
					delta = timedelta(minutes=span)
				elif rtype=='SECONDLY':
					delta = timedelta(seconds=span)
				elif rtype in ('MONTHLY-MONTHDAY','MONTHLY-WEEKDAY'):
					cls._sync_rep_end_from_occ_rrule(rtype)
					return
				else:
					# !! Don't know how to sync
					print('Warning: Sync for {} not implemented'.format(rtype), file=stderr)
					return
				edt = stdt+delta
				cls.wid_rep_enddt.set_date(edt)
		else: # occurrences determined by end date - use dateutil:rrule to calc
			try:
				fr = cls.MAP_RTYPE_TO_RRULE[rtype]
			except KeyError:
				# !! Don't know how to sync
				print('Warning: Sync for {} not implemented'.format(rtype), file=stderr)
				return
			bymtdy = None
			bywkdy = None
			if rtype=='MONTHLY-MONTHDAY':
				bymtdy = cls._get_cal_monthday_rep()
			elif rtype=='MONTHLY-WEEKDAY':
				bywkdy = cls._get_cal_weekday_rep()
			dtst = cls.get_date_start() # Without time, because time breaks calc
			interv = cls.get_repeat_interval()
			rend = cls.wid_rep_enddt.get_date_or_none()
			if dtst is not None and rend is not None:
				rr = du_rrule.rrule(fr, dtstart=dtst, interval=interv, until=rend, byweekday=bywkdy, bymonthday=bymtdy)
				c = rr.count()
				if c>=0:
					cls._set_occs_min(0 if c==0 else 1) # possibly allow "0"
					cls.wid_rep_occs.set_value(c)


	@classmethod
	def _sync_rep_end_from_occ_rrule(cls, rtype:str) -> None:
		# Function to synchronise End Date repeat field from Occurrences in
		# complex situations, e.g. leapday (29 Feb) or monthly late in month.
		# Use rrule class for these.
		try:
			fr = cls.MAP_RTYPE_TO_RRULE[rtype]
		except KeyError:
			# !! Don't know how to sync
			print('Warning: Sync for {} not implemented'.format(rtype), file=stderr)
			return
		dtst = cls.get_datetime_start()
		if dtst is None:
			return
		interv = cls.get_repeat_interval()
		occs = cls.get_repeat_occurrences()
		bymtdy = None
		bywkdy = None
		if rtype=='MONTHLY-MONTHDAY':
			bymtdy = cls._get_cal_monthday_rep()
		elif rtype=='MONTHLY-WEEKDAY':
			bywkdy = cls._get_cal_weekday_rep()
		rr = du_rrule.rrule(fr, dtstart=dtst, interval=interv, count=occs, byweekday=bywkdy, bymonthday=bymtdy)
		edt = list(rr)[-1]
		cls.wid_rep_enddt.set_date(edt)


	@classmethod
	def _get_cal_monthday_rep(cls) -> int:
		# Return value for monthday repeats to pass to rrule
		retd = int(cls.wid_repbymonthday.get_active_id())
		return retd


	@classmethod
	def _get_cal_weekday_rep(cls) -> du_rrule.weekday:
		# Return rrule structure for weekday repeats.
		# E.g. MO(-2) = "2nd last Monday"
		rrwd = cls.MAP_RDAY_TO_RRULEDAY[cls.wid_repbyweekday_day.get_active_id()](int(cls.wid_repbyweekday_ord.get_active_id()))
		return rrwd


	@classmethod
	def newentry(cls, txt:str=None) -> None:
		# Called to implement "new entry" from GUI, e.g. button
		cls.dialog.set_title(_('New Entry'))
		cls._empty_desc_allowed = True # initially allowed, can switch to False
		response,ei = cls._do_entry_dialog(txt=txt)
		if response==Gtk.ResponseType.OK and ei.desc:
			Calendar.new_entry(ei)
			if ei.rep_type is None:
				# Jump to entry date (not repeating, so well-defined)
				# !! This is a quick fix most common case. Need to fix:
				#    - Multiple entries on one day - jump to correct
				#    - Repeating entries (jump to visible/closest)
				GUI.cursor_date = ei.get_start_date()
			GUI.view_redraw(True)


	@classmethod
	def editentry(cls, en:iEvent, subtab:int=None) -> None:
		# Called to implement "edit entry" from GUI
		cls.dialog.set_title(_('Edit Entry'))
		cls._empty_desc_allowed = None # empty desc always allowed (for delete)
		response,ei = cls._do_entry_dialog(entry=en, subtab=subtab)
		if response==Gtk.ResponseType.OK:
			if ei.desc:
				Calendar.update_entry(en, ei)
				if ei.rep_type is None:
					# Jump to entry date
					GUI.cursor_date = ei.get_start_date()
				GUI.view_redraw(True)
			else: # Description text has been deleted in dialog
				GUI.dialog_deleteentry(en)


	@classmethod
	def _do_entry_dialog(cls, entry:iEvent=None, txt:str=None, subtab:int=None) -> Tuple[int,EntryInfo]:
		# Do the core work displaying entry dialog and extracting result.
		# Called from both newentry() and editentry().
		cls._seed_fields(entry, txt)

		# Select visible subtab: time, repeats, alarms...
		cls.wid_tabs.set_current_page(0) # default
		if subtab is not None: # jump to tab requested
			cls.wid_tabs.set_current_page(subtab)
			cls.wid_tabs.grab_focus()
			cls.wid_desc.select_region(0,0) # remove highlight

		cls._reset_err_style() # clear old error highlights

		# Unblock signals when dialog is active (user interaction -> updates)
		cls._unblock_tmdur_signals()
		cls._unblock_rep_occend_signals()
		try:
			while True:
				response = cls.dialog.run()
				if response!=Gtk.ResponseType.OK or cls._is_valid_entry(set_style=True):
					break
				cls._cancel_empty_desc_allowed() # after OK, desc required
		finally:
			# re-block signals
			cls._block_tmdur_signals()
			cls._block_rep_occend_signals()
			cls.dialog.hide()

		return response,cls._get_entryinfo()


	@classmethod
	def _seed_fields(cls, entry:iEvent, txt:Optional[str]) -> None:
		# Initialise entry fields when dialog is opened.
		# Data optionally from an entry, or text used as summary.
		# !! This function is a horrible mess. Needs rationalising !!

		# Set defaults
		dur = None
		end_dttm = None
		cls.dur_determines_end = True
		cls.wid_allday_count.set_value(1)
		cls.wid_rep_type.set_active(0)
		cls.repbymonthday_initialized = False # Init these later when we can
		cls.repbyweekday_initialized = False  # choose appropriate values.
		cls.wid_rep_interval.set_value(1)
		cls.wid_rep_forever.set_active(True)
		cls.rep_occs_determines_end = True
		cls._set_occs_min(1)
		cls.wid_rep_occs.set_value(1)
		# No need to set wid_rep_enddt because it will be synced when revealed
		cls.wid_status.set_active(0)
		cls.wid_location.set_text('')
		cls.wid_alarmset.set_active(False)

		# Two cases: new entry or edit existing entry
		if entry is None:
			# Initialise decscription field
			# We don't want this to count as user interaction, so block signal
			cls.wid_desc.handler_block(cls._wid_desc_changed_handler)
			if txt is None:
				cls.wid_desc.set_text('') # clear text
			else:
				cls.wid_desc.set_text(txt)
				cls.wid_desc.set_position(len(txt))
			cls.wid_desc.handler_unblock(cls._wid_desc_changed_handler)#unblock
			cls.wid_desc.grab_focus_without_selecting()
			dt = GUI.cursor_date
			tm = None
		else: # existing entry - take values
			cls.wid_desc.set_text(entry['SUMMARY'] if 'SUMMARY' in entry else '')
			cls.wid_desc.grab_focus()
			dt = entry['DTSTART'].dt
			dttm = None
			if isinstance(dt,dt_datetime):
				dttm = dt
				dt = dttm.date()
				tm = dttm.time()
				if 'DTEND' in entry:
					end_dttm = entry['DTEND'].dt
					if isinstance(end_dttm, dt_time):
						end_dttm = dt_datetime.combine(dt,end_dttm)
						if end_dttm<dttm:
							end_dttm += timedelta(days=1)
				elif 'DURATION' in entry:
					dur = entry['DURATION'].dt
			elif isinstance(dt,dt_date):
				tm = None
				if 'DTEND' in entry and isinstance(entry['DTEND'].dt, dt_date):
					end_dttm = entry['DTEND'].dt
			else: # dt is neither a date nor a datetime
				raise TypeError("Entry date of unexpected type")
			if 'RRULE' in entry:
				rrule = entry['RRULE']
				rrfreq = rrule['FREQ'][0]
				if rrfreq == 'MONTHLY' and 'BYDAY' in rrule:
					cls.wid_rep_type.set_active_id('MONTHLY-WEEKDAY')
					byday = rrule['BYDAY'][0]
					cls.repbyweekday_initialized = True
					cls.wid_repbyweekday_ord.set_active_id(byday[1 if byday[0]=='+' else 0:-2])
					cls.wid_repbyweekday_day.set_active_id(byday[-2:])
				elif rrfreq == 'MONTHLY' and 'BYMONTHDAY' in rrule:
					if len(rrule['BYMONTHDAY'])!=1:
						raise TypeError('Editing repeat with multiple \'BYMONTHDAY\' not (yet) supported')
					cls.wid_rep_type.set_active_id('MONTHLY-MONTHDAY')
					bymday = rrule['BYMONTHDAY'][0]
					if not(-7 <= int(bymday) <= -1):
						raise TypeError('Editing repeat with BYMONTHDAY={} not (yet) supported'.format(bymday))
					cls.wid_repbymonthday.set_active_id(str(bymday))
					cls.repbymonthday_initialized = True
				elif rrfreq in ('YEARLY','MONTHLY','WEEKLY','DAILY'):
					cls.wid_rep_type.set_active_id(rrfreq)
				else:
					raise TypeError('Editing repeat freq \'{}\' not (yet) supported'.format(rrfreq))
				cls.wid_rep_interval.set_value(int(rrule['INTERVAL'][0]) if 'INTERVAL' in rrule else 1)
				if 'COUNT' in rrule:
					cls.wid_rep_forever.set_active(False)
					c = rrule['COUNT'][0]
					cls.wid_rep_occs.set_value(c if c>=1 else 1)
					cls.rep_occs_determines_end = True
				elif 'UNTIL' in rrule:
					cls.wid_rep_forever.set_active(False)
					u = rrule['UNTIL'][0]
					if isinstance(u,dt_datetime):
						u = u.date()
					cls.wid_rep_enddt.set_date(u if u>dt else dt)
					cls.rep_occs_determines_end = False
			if 'STATUS' in entry and entry['STATUS'] in ('TENTATIVE','CONFIRMED','CANCELLED'):
				cls.wid_status.set_active_id(entry['STATUS'])
			if 'LOCATION' in entry:
				cls.wid_location.set_text(entry['LOCATION'])
			if entry.walk('VALARM'):
				cls.wid_alarmset.set_active(True)

		cls.wid_date.set_date(dt)
		cls._sync_rep_occs_end()

		if tm is None:
			# Setting radio buttons for Untimed/Timed/Allday.
			# This also reveals appropriate UI elements via signal connections.
			if end_dttm is None:
				cls.wid_timed_buttons[0].set_active(True)
			else:
				cls.wid_timed_buttons[2].set_active(True)
				d = end_dttm - dt
				cls.wid_allday_count.set_value(d.days)
			tm = dt_time(hour=9)
		else:
			cls.wid_timed_buttons[1].set_active(True)
		cls.wid_time.set_time(tm)

		if dur is not None:
			end_dttm = dttm + dur
			cls.dur_determines_end = True
		elif end_dttm is not None and dttm is not None:
			dur = end_dttm - dttm
			cls.dur_determines_end = False

		if dur is None:
			cls.wid_dur.set_duration(timedelta(0))
			cls.wid_endtime.set_time(tm)
		else:
			cls.wid_dur.set_duration(dur)
			cls.wid_endtime.set_time(end_dttm.time())

		cls._seed_rep_exception_list(entry)


	@classmethod
	def _seed_rep_exception_list(cls, entry:iEvent) -> None:
		# Sets & displays repeat exception list from entry parameter.
		# Called from _seed_fields() when dialog is opened.
		cls._set_rep_exception_list(entry)
		cls._set_label_rep_list()


	@classmethod
	def _set_rep_exception_list(cls, entry:iEvent) -> None:
		# Set cls.exception_list to list of exception dates for entry
		cls.exception_list = []
		if entry is None or 'RRULE' not in entry or 'EXDATE' not in entry:
			return
		exdate = entry['EXDATE']
		if isinstance(exdate,list):
			# Create set => elements unique
			dtlist = set([datetime_to_date(t.dts[i].dt) for t in exdate for i in range(len(t.dts))])
		else:
			dtlist = set([datetime_to_date(t.dt) for t in exdate.dts])
		cls.exception_list = sorted(dtlist)


	@classmethod
	def _set_label_rep_list(cls) -> None:
		# Construct string and set text for label widget displaying
		# exception dates. Takes dates from cls.exception_list.
		list_txt = ''
		first = True
		for dt in cls.exception_list:
			if not first:
				list_txt += ', '
			list_txt += dt.strftime(GUI.date_formatting_numeric)
			first = False
		if not list_txt:
			list_txt = _('None')
		cls._lab_rep_exceptions.set_text(list_txt)


	@classmethod
	def _get_entryinfo(cls) -> EntryInfo:
		# Decipher entry fields and return info as an EntryInfo object.
		desc = cls.wid_desc.get_text()
		dt = cls.get_datetime_start()
		stat = cls.wid_status.get_active_id()
		loc = cls.wid_location.get_text()
		ei = EntryInfo(desc=desc, start_dt=dt, status=stat, location=loc)
		if cls.wid_timed_buttons[2].get_active():
			# "Day entry" selected, read number of days from widget
			d = max(1,int(cls.wid_allday_count.get_value()))
			ei.set_end_dt(dt+timedelta(days=d))
		elif cls.dur_determines_end:
			ei.set_duration(cls.wid_dur.get_duration_or_none())
		else:
			ei.set_end_dt(cls.wid_endtime.get_time_or_none())

		# repeat information
		reptype = cls.wid_rep_type.get_active_id()
		if reptype is not None:
			inter = cls.get_repeat_interval()
			count = None
			until = None
			byday = None
			bymonthday = None
			if reptype=='MONTHLY-WEEKDAY':
				reptype = 'MONTHLY'
				byday = cls.wid_repbyweekday_ord.get_active_id() + cls.wid_repbyweekday_day.get_active_id()
			elif reptype=='MONTHLY-MONTHDAY':
				reptype = 'MONTHLY'
				bymonthday = cls.wid_repbymonthday.get_active_id()
			if not cls.wid_rep_forever.get_active():
				if cls.rep_occs_determines_end:
					count = cls.get_repeat_occurrences()
				else:
					until = cls.wid_rep_enddt.get_date_or_none()
			ei.set_repeat_info(reptype, interval=inter, count=count, until=until, bymonthday=bymonthday, byday=byday, except_list=cls.exception_list)
		return ei


	@classmethod
	def _is_valid_entry(cls, set_style:bool=False) -> bool:
		# Function checks if entry dialog fields are valid.
		# Used when "OK" clicked in the dialog (so important!).
		# Returns True if all fields valid; False if *any* field is invalid.
		# If an entry is invalid and set_style==True it also adds a
		# CSS class to the widget, so the error is visibly indicated.
		# If the entry is valid then it removes the error style, so
		# the error indication will disappear (regardless of set_style).

		ret = True
		# Check description is good (optional)
		ctx = cls.wid_desc.get_style_context()
		if cls._empty_desc_allowed==False and not cls.wid_desc.get_text(): # is empty string
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Check date is good (e.g. not February 30th)
		ctx = cls.wid_date.get_style_context()
		st_dt = cls.get_date_start()
		if st_dt is None:
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
		else: # always try to remove style, so corrections visible
			ctx.remove_class(GUI.STYLE_ERR)

		# Track tabs were errors found so they can be selected if needed
		err_tabs = [False]*cls.TAB_COUNT

		# Check start time is good (e.g. not 10:70)
		ctx = cls.wid_time.get_style_context()
		if cls.wid_time.get_time_or_none() is None:
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
				err_tabs[cls.TAB_TIME] = True
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Check duration is good
		# Don't want to do an if/else here because need to reset conterpart
		ctx = cls.wid_dur.get_style_context()
		if cls.dur_determines_end and cls.wid_dur.get_duration_or_none() is None:
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
				err_tabs[cls.TAB_TIME] = True
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Check end time is good
		ctx = cls.wid_endtime.get_style_context()
		if not cls.dur_determines_end and cls.wid_endtime.get_time_or_none() is None:
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
				err_tabs[cls.TAB_TIME] = True
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Shared Boolean used in checking repeat enddate & occs
		reps_active = cls.wid_rep_type.get_active()>0 and not cls.wid_rep_forever.get_active()

		# Check end repeat date is good
		ctx = cls.wid_rep_enddt.get_style_context()
		end_dt = cls.wid_rep_enddt.get_date_or_none()
		if reps_active and (end_dt is None or (st_dt is not None and end_dt<st_dt)):
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
				err_tabs[cls.TAB_REPEATS] = True
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Check repeat occurrences is good
		ctx = cls.wid_rep_occs.get_style_context()
		if reps_active and cls.wid_rep_occs.get_value()<1:
			ret = False
			if set_style:
				ctx.add_class(GUI.STYLE_ERR)
				err_tabs[cls.TAB_REPEATS] = True
		else:
			ctx.remove_class(GUI.STYLE_ERR)

		# Move to a tab with highlighted errors if not already on one
		if set_style and not err_tabs[cls.wid_tabs.get_current_page()]:
			for i in range(cls.TAB_COUNT):
				if err_tabs[i]:
					cls.wid_tabs.set_current_page(i)
					break

		return ret


	@classmethod
	def _reset_err_style(cls) -> None:
		# Remove 'dialog_error' style class from entry dialog widgets where
		# it might be set, so any visible indications of errors are removed.
		# Used, for example, when the dialog is reinitialised.
		for w in (cls.wid_desc, cls.wid_date, cls.wid_time, cls.wid_dur, cls.wid_endtime, cls.wid_rep_enddt, cls.wid_rep_occs):
			w.get_style_context().remove_class(GUI.STYLE_ERR)


	@classmethod
	def get_date_start(cls) -> dt_date:
		# Convenience function to get start date from dialog (None if invalid)
		return cls.wid_date.get_date_or_none()


	@classmethod
	def get_datetime_start(cls) -> dt_date:
		# Returns start date & time from dialog if time set; o/w just the date
		dt = cls.wid_date.get_date_or_none()
		if dt is not None and cls.wid_timed_buttons[1].get_active():
			# timed entry
			tm = cls.wid_time.get_time_or_none()
			if tm is not None:
				dt = dt_datetime.combine(dt,tm)
		return dt


	@classmethod
	def get_repeat_interval(cls) -> int:
		# Use get_text() rather than get_value() since there can be
		# delays in value being updated. This can lead to outdated
		# values being reported. This happens almost all the time
		# on the Gemini device.
		try:
			r = int(cls.wid_rep_interval.get_text())
			if r<1:
				r = 1
		except ValueError:
			r = 1
		return r


	@classmethod
	def get_repeat_occurrences(cls) -> int:
		# Use get_text() rather than get_value() - see above.
		try:
			r = int(cls.wid_rep_occs.get_text())
			if r<1:
				r = 1
		except ValueError:
			r = 1
		return r


	@classmethod
	def dialog_repeat_exceptions(cls, *args) -> None:
		# Callback for +/- button to open sub-dialog for "repeat exceptions"
		edc = ExceptionsDialogController(cls.exception_list, parent=cls.dialog)
		result = edc.run()
		if result==Gtk.ResponseType.OK:
			dates = sorted(edc.get_dates())
			cls.exception_list = dates
			cls._set_label_rep_list()
		edc.destroy()


class DateLabel(Gtk.Label):
	# Label Widget displaying a date.
	# Used for rows in listbox of Exception dates dialog.

	def __init__(self, date:dt_date):
		# !! date argument assumed a dt_date here; may be dt_datetime in future
		self.date = date
		st = self.date.strftime(GUI.date_formatting_textabb)
		super().__init__(st)


	def compare(self, dl2:Gtk.Label) -> int:
		# Compare two DateLabels.
		# Return <0 if self<dl2; 0 if equal; >0 otherwise.
		# Used to sort ListBox.
		# !! If using datetime, then this won't distinguish within day
		o1 = self.date.toordinal()
		o2 = dl2.date.toordinal()
		return o1-o2


class ExceptionsDialogController:
	# Dialog box to add/remove a repeating entry's "Exception dates".
	# Called when +/- button is clicked in Entry Repeats tab.

	def __init__(self, ex_date_list:list, parent:Gtk.Window=None):
		self.dialog = Gtk.Dialog(title=_('Exception dates'), parent=parent,
		    flags=Gtk.DialogFlags.MODAL|Gtk.DialogFlags.DESTROY_WITH_PARENT,
		    buttons=(Gtk.STOCK_CANCEL, Gtk.ResponseType.CLOSE, Gtk.STOCK_OK, Gtk.ResponseType.OK))
		self.dialog.set_resizable(False)
		content = Gtk.Box()
		add_content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		list_content = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
		content.add(add_content)
		content.add(list_content)
		self.dialog.get_content_area().add(content)

		# Date widget
		self.wdate = WidgetDate(GUI.cursor_date)
		self.wdate.connect('changed', GUI.check_date_fixed)
		add_content.add(self.wdate)

		# Add date button
		button_add = Gtk.Button(_('Add date'))
		button_add.get_style_context().add_class('dialogbutton')
		button_add.connect('clicked', self._add_date, self)
		add_content.add(button_add)

		# Date listbox (scrollable)
		list_scroller = Gtk.ScrolledWindow()
		list_scroller.set_size_request(-1,70) # temporary magic number !!
		list_scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)
		list_scroller.set_overlay_scrolling(False)
		self.list_listbox = Gtk.ListBox()
		self.list_listbox.set_selection_mode(Gtk.SelectionMode.MULTIPLE)
		self.list_listbox.set_activate_on_single_click(False)
		self.list_listbox.set_sort_func(self._sort_func)
		for dt in set(ex_date_list): # set() to uniquify
			lab = DateLabel(dt)
			self.list_listbox.add(lab)
		list_scroller.add(self.list_listbox)
		list_content.add(list_scroller)

		# Remove date(s) button
		button_remove = Gtk.Button(_('Remove date(s)'))
		button_remove.get_style_context().add_class('dialogbutton')
		button_remove.connect('clicked', self._remove_dates, self)
		list_content.add(button_remove)

		# !! Add WIP notice, so status is clear to user !!
		wip_label = Gtk.Label('Note: this dialog needs refining!')
		self.dialog.get_content_area().add(wip_label)

		self.dialog.show_all()


	def run(self) -> Gtk.ResponseType:
		# run the dialog
		return self.dialog.run()


	def destroy(self) -> None:
		# destroy the dialog
		self.dialog.destroy()


	def get_dates(self) -> set:
		# Return list of dates.
		# Returned as a set, so guaranteed no repeats, but not sorted.
		dl = {row.get_child().date for row in self.list_listbox.get_children()}
		return dl


	@staticmethod
	def _add_date(button:Gtk.Button, ctrl:'ExceptionsDialogController') -> None:
		# Callback when "Add date" button is clicked
		dt = ctrl.wdate.get_date_or_none()
		if dt is None:
			# Date is invalid, add error styling
			ctrl.wdate.get_style_context().add_class(GUI.STYLE_ERR)
		else:
			ctrl.list_listbox.unselect_all() # so new row can be highlighted
			for row in ctrl.list_listbox.get_children():
				if dt == row.get_child().date:
					# date already in list - highlight & stop
					ctrl.list_listbox.select_row(row)
					return
			lab = DateLabel(dt)
			lab.show_all()
			ctrl.list_listbox.add(lab)
			# Now select the new row - helps to see where new entry is,
			# and also means it can be quickly deleted if added in error.
			ctrl.list_listbox.select_row(lab.get_parent())


	@staticmethod
	def _remove_dates(button:Gtk.Button, ctrl:'ExceptionsDialogController') -> None:
		# Callback when "Remove dates" button is clicked
		selected = ctrl.list_listbox.get_selected_rows()
		for row in selected:
			ctrl.list_listbox.remove(row)


	@staticmethod
	def _sort_func(row1:Gtk.ListBoxRow, row2:Gtk.ListBoxRow) -> int:
		# Used as sort function for listbox.
		return row1.get_child().compare(row2.get_child())



# -*- coding: utf-8 -*-
#
# pygenda_view_week.py
# Provides the "Week View" for Pygenda.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from gi import require_version as gi_require_version
gi_require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk, GLib

from calendar import day_abbr,month_name
from datetime import time as dt_time, date as dt_date, timedelta
from icalendar import cal as iCal
from locale import gettext as _
from typing import Optional

# pygenda components
from .pygenda_view import View
from .pygenda_calendar import Calendar
from .pygenda_config import Config
from .pygenda_util import start_of_week, day_in_week, month_abbr, start_end_dts_occ, dt_lte
from .pygenda_gui import GUI, EntryDialogController


# Singleton class for Week View
class View_Week(View):
	Config.set_defaults('week_view',{
		'pageleft_datepos': 'left',
		'pageright_datepos': 'right',
	})

	_day_ent_count = [0]*7 # entry count for each day
	_visible_occurrences = []
	_week_viewed = None # So view will be fully redrawn when needed
	_last_cursor = None
	_scroll_callback_id = None
	CURSOR_STYLE = 'weekview_cursor'

	@staticmethod
	def view_name() -> str:
		# Return (localised) string to use in menu
		return _('_Week View')

	@staticmethod
	def accel_key() -> int:
		# Return (localised) keycode for menu shortcut
		k = _('week_view_accel')
		return ord(k[0]) if len(k)>0 else 0


	@classmethod
	def init(cls) -> Gtk.Widget:
		# Called on startup.
		# Gets view framework from glade file & tweaks/adds a few elements.
		# Returns widget containing view.
		cls._topbox = GUI._builder.get_object('view_week')
		cls._month_label = GUI._builder.get_object('week_label_month')
		cls._weekno_label = GUI._builder.get_object('week_label_weekno')
		cls._init_week_widgets()
		cls._init_keymap()
		return cls._topbox


	@classmethod
	def _init_week_widgets(cls) -> None:
		# Initialise widgets - create day labels, entry spaces etc.
		# Do this here so it take account of start_week_day setting,
		# page*_datepos settings, and set CSS styles for each day.

		# Create widgets for day labels & text spaces
		cls._day_eventbox = []
		cls._day_label = []
		cls._day_rows = []
		cls._day_scroll = []
		st_wk = Config.get_int('global','start_week_day')
		dpos_r = Config.get('week_view','pageleft_datepos')=='right'
		day_ab = ('mon','tue','wed','thu','fri','sat','sun')#don't trans
		for i in range(7):
			# create event box & contained box
			cls._day_eventbox.append(Gtk.EventBox())
			day_box = Gtk.Box()
			ctx = day_box.get_style_context()
			# Add classes for css flexibility
			ctx.add_class('weekview_day')
			ctx.add_class('weekview_day{:d}'.format(i))
			ctx.add_class('weekview_{:s}'.format(day_ab[(i+st_wk)%7]))
			cls._day_eventbox[i].add(day_box)
			cls._day_eventbox[i].connect("button_press_event", cls.click_date)
			# labels
			cls._day_label.append(Gtk.Label())
			cls._day_label[i].set_justify(Gtk.Justification.CENTER)
			cls._day_label[i].set_xalign(0.5)
			cls._day_label[i].set_yalign(0.5)
			ctx = cls._day_label[i].get_style_context()
			ctx.add_class('weekview_labelday')
			if i==3: # starting rightpage
				dpos_r = Config.get('week_view','pageright_datepos')=='right'
			if dpos_r:
				day_box.pack_end(cls._day_label[i], False, False, 0)
			else:
				day_box.pack_start(cls._day_label[i], False, False, 0)
			# scroller & textview
			day_scroller = Gtk.ScrolledWindow()
			cls._day_scroll.append(day_scroller)
			day_scroller.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
			day_scroller.set_overlay_scrolling(False)
			day_scroller.set_hexpand(True)
			cls._day_rows.append(Gtk.Box(orientation=Gtk.Orientation.VERTICAL))
			day_scroller.add(cls._day_rows[i])
			ctx = cls._day_rows[i].get_style_context()
			ctx.add_class('weekview_daytext')
			day_box.pack_start(day_scroller, True, True, 0)

		# Attach elements to pages
		page_l = GUI._builder.get_object('week_page_l')
		page_r = GUI._builder.get_object('week_page_r')
		for i in range(3): # Left
			page_l.pack_start(cls._day_eventbox[i], True, True, 0)
		for i in range(3,7): # ... and right
			page_r.pack_start(cls._day_eventbox[i], True, True, 0)


	@classmethod
	def _set_label_text(cls) -> None:
		# Sets date and month label text and style classes.
		# Called on view redraw.
		cls._last_cursor = None
		dt = start_of_week(GUI.cursor_date)
		cls._week_viewed = dt
		dt_end = dt + timedelta(days=6)
		# Month label
		if dt.month == dt_end.month:
			mn = month_name[dt.month].capitalize()
			ml = u'{mon:s} {yr:d}'.format(mon=mn, yr=dt.year)
		elif dt.year == dt_end.year:
			msn = month_abbr[dt.month].capitalize()
			ml = u'{mon_st:s} – {mon_end:s} {yr:d}'.format(mon_st=msn, mon_end=month_abbr[dt_end.month], yr=dt.year)
		else: # month & year different
			msn = month_abbr[dt.month].capitalize()
			ml = u'{mon_st:s} {yr_st:d} – {mon_end:s} {yr_end:d}'.format(mon_st=msn, mon_end=month_abbr[dt_end.month], yr_st=dt.year, yr_end=dt_end.year)
		cls._month_label.set_text(ml)

		# Week number label
		# Week 1 is the first to include 4+ days of year.
		# E.g. if week starts Monday then Wk1 is first to include a Thursday.
		wkno = (dt.timetuple().tm_yday+9)//7
		if wkno==53 and dt.day>=29:
			wkno_txt = _('Week 53/Week 1')
		else:
			wkno_txt = _('Week {}').format(wkno)
		cls._weekno_label.set_text(wkno_txt)

		# Day labels
		today = dt_date.today()
		for i in range(7):
			ctx = cls._day_eventbox[i].get_style_context()
			if dt<today:
				ctx.add_class('weekview_pastday')
			else:
				ctx.remove_class('weekview_pastday')
			if dt==today:
				ctx.add_class('weekview_today')
			else:
				ctx.remove_class('weekview_today')

			dn = day_abbr[dt.weekday()]
			if dn[-1]=='.':
				dn = dn[:-1] # remove dot
			cls._day_label[i].set_text(u'{day:s}\n{date:d}'.format(day=dn,date=dt.day))
			dt += timedelta(days=1)


	@classmethod
	def _set_entry_text(cls) -> None:
		# Sets label text and style classes for event-displaying labels.
		# Called on view redraw.
		cls._last_cursor = None
		dt = start_of_week(GUI.cursor_date)
		cls._visible_occurrences = Calendar.occurrence_list(dt, dt+timedelta(days=7))
		itr = iter(cls._visible_occurrences)
		try:
			occ = next(itr)
		except StopIteration:
			occ = None
		cls._day_ent_count = [0]*7 # we'll fill this in as we display
		for i in range(7):
			dt_nxt = dt + timedelta(days=1)
			# Delete anything previously written to day v-box
			cls._day_rows[i].foreach(Gtk.Widget.destroy)
			while True:
				if occ is None:
					break
				occ_dt_sta,occ_dt_end = start_end_dts_occ(occ)
				if dt_lte(dt_nxt, occ_dt_sta):
					# into next day so break this loop
					break
				row = Gtk.Box()
				# Create entry mark (bullet or time) & add to row
				mark_label = cls.marker_label(occ[0], occ_dt_sta)
				ctx = mark_label.get_style_context()
				ctx.add_class('weekview_marker') # add style for CSS
				row.add(mark_label)

				# Create entry content label & add to row
				cont_label = cls.entry_text_label(occ[0],occ_dt_sta,occ_dt_end)
				cont_label.set_hexpand(True) # Also sets hexpand_set to True
				row.add(cont_label)
				cls._day_rows[i].add(row)
				cls._day_ent_count[i] += 1
				try:
					occ = next(itr)
				except StopIteration:
					occ = None
			if cls._day_ent_count[i]==0:
				# an empty day, need something for cursor
				mark_label = Gtk.Label()
				ctx = mark_label.get_style_context()
				ctx.add_class('weekview_marker')
				mark_label.set_text(u' ') # en-space char
				mark_label.set_halign(Gtk.Align.START) # else cursor fills line
				cls._day_rows[i].add(mark_label)
			dt = dt_nxt
			cls._day_rows[i].show_all()


	@classmethod
	def _show_cursor(cls) -> None:
		# Locates bullet/date corresponding to the current cursor and adds
		# 'weekview_cursor' class to it so cursor is visible via CSS styling.
		dy = day_in_week(GUI.cursor_date)
		ecount = cls._day_ent_count[dy]
		i = GUI.cursor_idx_in_date
		if i < 0 or i >= ecount:
			i = max(0,ecount-1)
			GUI.cursor_idx_in_date = i
		cls._hide_cursor()
		row = cls._day_rows[dy].get_children()[i]
		if ecount==0:
			mk = row
		else:
			mk = row.get_children()[0]
		ctx = mk.get_style_context()
		ctx.add_class(cls.CURSOR_STYLE)
		cls._last_cursor = int(dy+8*i)
		if cls._scroll_callback_id is not None:
			# Cancel existing callback to prevent inconsistent scroll requests.
			GLib.source_remove(cls._scroll_callback_id)
			cls._scroll_callback_id = None
		if cls._day_ent_count[dy] > 0:
			# We may need to scroll content to show entry at cursor.
			# Want this to happen after redraw, otherwise, can't access
			# dimensions of elements needed to calculate scroll size (because
			# they haven't been calculated). So delay scroll to after redraw.
			# We save the returned id so the callback can be cancelled.
			cls._scroll_callback_id = GLib.idle_add(cls._scroll_to_row, cls._day_rows[dy], i, cls._day_scroll[dy], priority=GLib.PRIORITY_HIGH_IDLE+30)


	@classmethod
	def _scroll_to_row(cls, rowbox:Gtk.Box, row:int, scroller:Gtk.ScrolledWindow) -> bool:
		# Local version of scroll_to_row() that clears the id, then
		# calls the parent class scroll_to_row() to do the work.
		cls._scroll_callback_id = None
		return cls.scroll_to_row(rowbox, row, scroller)


	@classmethod
	def _hide_cursor(cls) -> None:
		# Clears 'weekview_cursor' style class from cursor position,
		# so cursor is no longer visible.
		if cls._last_cursor is not None:
			# _last_cursor is an int split into two parts:
			# Lower 3 bits give day, other higher bits give entry within day
			dy = cls._last_cursor%8
			row = cls._day_rows[dy].get_children()[cls._last_cursor//8]
			if cls._day_ent_count[dy]==0:
				mk = row
			else:
				mk = row.get_children()[0]
			ctx = mk.get_style_context()
			ctx.remove_class(cls.CURSOR_STYLE)
			cls._last_cursor = None


	@classmethod
	def get_cursor_entry(cls) -> iCal.Event:
		# Returns entry at cursor position, or None if cursor not on entry.
		# Called from cursor_edit_entry().
		dy = day_in_week(GUI.cursor_date)
		if cls._day_ent_count[dy]==0:
			return None
		i = sum(cls._day_ent_count[:dy])
		i += GUI.cursor_idx_in_date
		return cls._visible_occurrences[i][0]


	@classmethod
	def renew_display(cls) -> None:
		# Called when we switch to this view to reset state.
		cls._week_viewed = None


	@classmethod
	def redraw(cls, ev_changes:bool) -> None:
		# Called when redraw required.
		# ev_changes: bool indicating if events need updating too
		if cls._week_viewed != start_of_week(GUI.cursor_date):
			cls._set_label_text()
			ev_changes = True
		if ev_changes:
			cls._set_entry_text()
		cls._show_cursor()


	@classmethod
	def _cursor_move_up(cls) -> None:
		# Callback for user moving cursor up.
		GUI.cursor_idx_in_date -= 1
		if GUI.cursor_idx_in_date < 0:
			GUI.cursor_inc(timedelta(days=-1))
			# Leave idx_in_date as -1 since that signals last entry
			GUI.today_toggle_date = None
		else:
			cls._show_cursor()


	@classmethod
	def _cursor_move_dn(cls) -> None:
		# Callback for user moving cursor down.
		GUI.cursor_idx_in_date += 1
		dy = day_in_week(GUI.cursor_date)
		if GUI.cursor_idx_in_date >= cls._day_ent_count[dy]:
			GUI.cursor_inc(timedelta(days=1), 0)
			GUI.today_toggle_date = None
		else:
			cls._show_cursor()


	@classmethod
	def _cursor_move_lt(cls) -> None:
		# Callback for user moving cursor left.
		i = day_in_week(GUI.cursor_date)
		d = -7 if i<3 else (-3 if i==3 else -4)
		GUI.cursor_inc(timedelta(days=d), 0)
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_rt(cls) -> None:
		# Callback for user moving cursor right.
		i = day_in_week(GUI.cursor_date)
		d = 4 if i<3 else 7
		GUI.cursor_inc(timedelta(days=d), 0)
		GUI.today_toggle_date = None


	@classmethod
	def _cursor_move_today(cls) -> None:
		# Callback for user toggling cursor between today & another date
		today = dt_date.today()
		if GUI.cursor_date != today:
			GUI.today_toggle_date = GUI.cursor_date
			GUI.today_toggle_idx = GUI.cursor_idx_in_date
			GUI.cursor_set(today,0)
		elif GUI.today_toggle_date is not None:
			GUI.cursor_set(GUI.today_toggle_date,GUI.today_toggle_idx)


	@classmethod
	def _init_keymap(cls) -> None:
		# Initialises KEYMAP for class. Called from init() since it needs
		# to be called after class construction, so that functions exist.
		cls._KEYMAP = {
			Gdk.KEY_Up: lambda: cls._cursor_move_up(),
			Gdk.KEY_Down: lambda: cls._cursor_move_dn(),
			Gdk.KEY_Right: lambda: cls._cursor_move_rt(),
			Gdk.KEY_Left: lambda: cls._cursor_move_lt(),
			Gdk.KEY_space: lambda: cls._cursor_move_today(),
			Gdk.KEY_Return: lambda: cls.cursor_edit_entry(),
		}


	@classmethod
	def keypress(cls, wid:Gtk.Widget, ev:Gdk.EventKey) -> None:
		# Handle key press event in Week view.
		# Called (from GUI.keypress()) on keypress (or repeat) event
		try:
			f = cls._KEYMAP[ev.keyval]
			GLib.idle_add(f)
		except KeyError:
			# If it's a character key, take as first of new entry
			# !! Bug: only works for ASCII characters
			if ev.state & (Gdk.ModifierType.CONTROL_MASK|Gdk.ModifierType.MOD1_MASK)==0 and Gdk.KEY_exclam <= ev.keyval <= Gdk.KEY_asciitilde:
				GLib.idle_add(EntryDialogController.newentry,chr(ev.keyval))


	@classmethod
	def click_date(cls, wid:Gtk.Widget, ev:Gdk.EventButton) -> None:
		# Callback. Called whenever a date is clicked/tapped.
		# Moves cursor to date/item clicked
		try:
			new_day = cls._day_eventbox.index(wid)
		except ValueError:
			return

		new_idx = 0 # default index if not going to specific entry

		# Has user clicked on a specific entry within day?
		# If so, move the cursor to that entry.
		if cls._day_ent_count[new_day] > 1: # Only check if >1 entry in the day
			# Check if clicked in day's label area
			d_wids = wid.get_child().get_children()
			ll = isinstance(d_wids[0],Gtk.Label) # True if left label
			if ll != (ev.x < d_wids[0].get_allocated_width()): # Clicked entries
				# We're not clicking in date label area, use y to calc entry
				new_idx = cls.y_to_day_row(cls._day_rows[new_day], ev.y, cls._day_ent_count[new_day], cls._day_scroll[new_day])

		GLib.idle_add(GUI.cursor_set, cls._day_index_to_date(new_day), new_idx)


	@classmethod
	def _day_index_to_date(cls, idx:int) -> Optional[dt_date]:
		# Given an day index idx=0...6 of day cell in the visible Week View,
		# return the corresponding date
		if cls._week_viewed is None:
			return None
		return cls._week_viewed+timedelta(days=idx)



# -*- coding: utf-8 -*-
#
# pygenda_calendar.py
# Connects to agenda data provider - either an ics file or CalDAV server.
#
# Copyright (C) 2022 Matthew Lewis
#
# This file is part of Pygenda.
#
# Pygenda is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# Pygenda is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pygenda. If not, see <https://www.gnu.org/licenses/>.


from icalendar import Calendar as iCalendar, Event as iEvent, vRecur
from datetime import timedelta, datetime as dt_datetime, time as dt_time, date as dt_date, timezone
from dateutil.relativedelta import relativedelta
from dateutil.rrule import rrulestr
from sys import stderr
from uuid import uuid1
from pathlib import Path
from functools import reduce
from os import stat as os_stat, chmod as os_chmod, rename as os_rename, path as os_path
import stat
from time import monotonic as time_monotonic
import tempfile
from typing import Optional

# Pygenda components
from .pygenda_config import Config
from .pygenda_util import dt_lt, dt_lte, datetime_to_date
from .pygenda_entryinfo import EntryInfo


# Interface base class to connect to different data sources.
# Used by Calendar class (below).
class CalendarConnector:
	def add_entry(self, event:iEvent) -> None:
		# Add a new entry component to the calendar data and store it.
		print('Warning: Add entry not implemented', file=stderr)

	def update_entry(self, event:iEvent) -> None:
		# Update an entry component in the calendar data and store it.
		print('Warning: Update not implemented', file=stderr)

	def delete_entry(self, event:iEvent) -> None:
		# Delete entry component to the calendar data and remove from store.
		print('Warning: Delete not implemented', file=stderr)


# Singleton class for calendar data access/manipulation
class Calendar:
	_entry_norep_list_sorted = None
	_entry_rep_list = None

	@classmethod
	def init(cls) -> None:
		# Calendar connector initialisation.
		# Can take a long time, since it loads/sorts calendar data.
		# Best called in background after GUI is started, or startup can be slow.
		Config.set_defaults('calendar',{
			'ics_file': None,
			'caldav_server': None,
			'caldav_username': None,
			'caldav_password': None,
			'caldav_calendar': None
		})

		# Read config values to get type of data source connector
		filename = Config.get('calendar','ics_file')
		caldav_server = Config.get('calendar','caldav_server')
		if caldav_server is not None:
			# Use either server or file, not both (at least for now)
			assert(filename is None)
			user = Config.get('calendar','caldav_username')
			passwd = Config.get('calendar','caldav_password')
			calname = Config.get('calendar','caldav_calendar')
			cls.calConnector = CalendarConnectorCalDAV(caldav_server,user,passwd,calname)
		else:
			# If no server given, use an ics file.
			# Use either the provided filename or a default name.
			if filename is None:
				filename = '{}/{}'.format(Config.conf_dirname,Config.DEFAULT_ICS_FILENAME)
			else:
				# Expand '~' (so it can be used in config file)
				filename =  os_path.expanduser(filename)
			# Create a connector for that file
			cls.calConnector = CalendarConnectorICSfile(filename)

		if cls.calConnector.cal.is_broken:
			print('Warning: Non-conformant ical file', file=stderr)


	@staticmethod
	def gen_uid() -> str:
		# Generate a UID for iCal elements (required element)
		uid = 'Pygenda-{}'.format(uuid1())
		return uid


	@classmethod
	def new_entry(cls, e_inf:EntryInfo) -> None:
		# Add a new iCal entry with content from event info object
		ev = iEvent()
		ev.add('UID', Calendar.gen_uid()) # Required
		# DateTime utcnow() function doesn't include TZ, so use now(tz.utc)
		utcnow = dt_datetime.now(timezone.utc)
		ev.add('DTSTAMP', utcnow) # Required
		ev.add('CREATED', utcnow) # Optional
		ev.add('SUMMARY', e_inf.desc)
		ev.add('DTSTART', e_inf.start_dt)
		cls._event_add_end_dur_from_info(ev,e_inf)

		# Repeats
		if e_inf.rep_type is not None and e_inf.rep_inter>0:
			cls._event_add_repeat_from_info(ev,e_inf)
			cls._entry_rep_list = None # Clear rep cache as modified
		else:
			cls._entry_norep_list_sorted = None # Clear norep cache as modified

		cls._event_set_status_from_info(ev, e_inf)
		cls._event_set_location_from_info(ev, e_inf)

		cls.calConnector.add_entry(ev) # Write to store


	@classmethod
	def new_entry_from_example(cls, exev:iEvent, dt_start:dt_date=None) -> None:
		# Add a new iCal entry to store given an iEvent as a "template".
		# Replace UID, timestamp etc. to make it a new event.
		# Potentially override exev's date/time with a new one.
		# Use to implement pasting events into new days/timeslots.
		ev = iEvent()
		ev.add('UID', Calendar.gen_uid()) # Required
		utcnow = dt_datetime.now(timezone.utc)
		ev.add('DTSTAMP', utcnow) # Required
		# Since it has a new UID, we consider it a new entry
		ev.add('CREATED', utcnow) # Optional
		summ = exev['SUMMARY'] if 'SUMMARY' in exev else None
		if not summ:
			summ = 'New entry' # fallback summary
		ev.add('SUMMARY', summ)
		ex_dt_start = exev['DTSTART'].dt if 'DTSTART' in exev else None
		if dt_start:
			if ex_dt_start:
				new_dt_start = ex_dt_start.replace(year=dt_start.year,month=dt_start.month,day=dt_start.day)
			else:
				new_dt_start = dt_start
		elif ex_dt_start:
			new_dt_start = ex_dt_start
		else:
			raise ValueError('Entry has no date/time')
		ev.add('DTSTART', new_dt_start)
		if 'DURATION' in exev:
			ev.add('DURATION', exev['DURATION'])
		elif ex_dt_start and 'DTEND' in exev:
			ex_dt_end = exev['DTEND'].dt
			if isinstance(ex_dt_start,dt_datetime) and (ex_dt_start.tzinfo is None or ex_dt_end.tzinfo is None):
				# If one tzinfo is None, make both None, so subtraction work
				ex_dt_start = ex_dt_start.replace(tzinfo=None)
				ex_dt_end = ex_dt_end.replace(tzinfo=None)
			dur = ex_dt_end - ex_dt_start
			ev.add('DTEND', new_dt_start + dur)
		if 'LOCATION' in exev:
			ev.add('LOCATION', exev['LOCATION'])
		cls.calConnector.add_entry(ev) # Write to store
		cls._entry_norep_list_sorted = None # Clear norep cache as modified


	@classmethod
	def update_entry(cls, ev:iEvent, e_inf:EntryInfo) -> None:
		# Update event using details from EntryInfo e_inf.
		clear_rep = False
		clear_norep = False

		if 'UID' not in ev:
			ev.add('UID', Calendar.gen_uid()) # Should be present
		# DateTime utcnow() function doesn't include TZ, so use now(tz.utc)
		utcnow =  dt_datetime.now(timezone.utc)
		try:
			ev['DTSTAMP'].dt = utcnow
		except KeyError:
			# Entry had no DTSTAMP (note: DTSTAMP required by icalendar spec)
			ev.add('DTSTAMP', utcnow)
		try:
			ev['LAST-MODIFIED'].dt = utcnow
		except KeyError:
			# Entry had no LAST-MODIFIED - add one
			ev.add('LAST-MODIFIED', utcnow)
		try:
			ev['SUMMARY'] = e_inf.desc
		except KeyError:
			# Entry had no SUMMARY
			ev.add('SUMMARY', e_inf.desc)

		# DTSTART - delete & re-add so type (DATE vs. DATE-TIME) is correct
		# (Also, Q: if comparing DTSTARTs with different TZs, how does != work?)
		if 'DTSTART' in ev:
			del(ev['DTSTART'])
		ev.add('DTSTART', e_inf.start_dt)

		# Duration or Endtime - first delete existing
		if 'DURATION' in ev:
			del(ev['DURATION'])
		if 'DTEND' in ev:
			del(ev['DTEND'])
		# Then add new end time/duration (if needed)
		cls._event_add_end_dur_from_info(ev,e_inf)

		# Repeats (including exception dates)
		if 'RRULE' in ev:
			del(ev['RRULE'])
			clear_rep = True
		else:
			clear_norep = True
		if 'EXDATE' in ev:
			del(ev['EXDATE'])
		if e_inf.rep_type is not None and e_inf.rep_inter>0:
			cls._event_add_repeat_from_info(ev,e_inf)
			clear_rep = True
		else:
			clear_norep = True

		# Other properties: status (cancelled, tentative, etc.), location
		cls._event_set_status_from_info(ev, e_inf)
		cls._event_set_location_from_info(ev, e_inf)

		# This needs optimising - some cases cause too much cache flushing !!
		if clear_norep:
			cls._entry_norep_list_sorted = None
		if clear_rep:
			cls._entry_rep_list = None

		cls.calConnector.update_entry(ev) # Write to store


	@staticmethod
	def _event_add_end_dur_from_info(ev:iEvent, e_inf:EntryInfo) -> None:
		# Adds end time or duration to an event from EntryInfo.
		# How it does this depends on whether event is timed or not

		# If entry is timed, check for an end-time/duration & add if present
		# Note: don't add both an end-time & duration - at most one
		if isinstance(e_inf.start_dt, dt_datetime):
			if e_inf.end_dt is not None:
				if isinstance(e_inf.end_dt, dt_datetime):
					if e_inf.end_dt > e_inf.start_dt:#need end_dt after start_dt
						ev.add('DTEND', e_inf.end_dt)
				elif isinstance(e_inf.end_dt, dt_time):
					end_dttm = dt_datetime.combine(e_inf.start_dt.date(),e_inf.end_dt)
					if end_dttm != e_inf.start_dt: # require end time to be after start time
						if end_dttm < e_inf.start_dt:
							end_dttm += timedelta(days=1)
						ev.add('DTEND', end_dttm)
			elif e_inf.duration is not None and isinstance(e_inf.duration, timedelta):
				if e_inf.duration.total_seconds()>0: # require duration > 0
					ev.add('DURATION', e_inf.duration)
		elif isinstance(e_inf.start_dt, dt_date) and isinstance(e_inf.end_dt, dt_date):
			# start & end are both dates (not times) => this is a day entry
			if e_inf.end_dt > e_inf.start_dt: # sanity check
				ev.add('DTEND', e_inf.end_dt)


	@staticmethod
	def _event_add_repeat_from_info(ev:iEvent, e_inf:EntryInfo) -> None:
		# Adds FREQ and EXDATE fields to event.
		# Assume these fields are empty (so either it's a
		# new entry or the delete happens elsewhere).
		rr_options = {'FREQ':[e_inf.rep_type]}
		if e_inf.rep_inter != 1:
			rr_options['INTERVAL'] = [e_inf.rep_inter]
		if e_inf.rep_until is not None:
			if isinstance(e_inf.start_dt,dt_datetime) and not isinstance(e_inf.rep_until,dt_datetime):
				# If start is a datetime, until needs to be a datetime
				dt_until = dt_datetime.combine(e_inf.rep_until,e_inf.start_dt.time())
			else:
				dt_until = e_inf.rep_until
			rr_options['UNTIL'] = [dt_until]
		elif e_inf.rep_count is not None:
			rr_options['COUNT'] = [e_inf.rep_count]
		if e_inf.rep_byday is not None:
			rr_options['BYDAY'] = [e_inf.rep_byday]
		if e_inf.rep_bymonthday is not None:
			rr_options['BYMONTHDAY'] = [e_inf.rep_bymonthday]
		ev.add('RRULE', rr_options)

		# Add exception date/times
		if e_inf.rep_exceptions:
			e_prm = {'VALUE':'DATE'} # ? Bug in icalendar 4.0.9 means we need to add VALUE parameter
			# Adding multiple EXDATE fields seems to be most compatible
			for rex in e_inf.rep_exceptions:
				ev.add('EXDATE', rex, parameters=e_prm)

		# For repeats, spec says DTSTART should correspond to RRULE
		# https://icalendar.org/iCalendar-RFC-5545/3-8-5-3-recurrence-rule.html
		# So check DTSTART is consistent with RRULE, and possibly adjust.
		# This ignores exceptions, since they are an extra layer.
		fst_occ = first_occ(ev['RRULE'].to_ical().decode('utf-8'),e_inf.start_dt)
		if fst_occ!=e_inf.start_dt:
			print('Notice: rewriting DTSTART to be consistent with RRULE', file=stderr)
			del(ev['DTSTART'])
			ev.add('DTSTART', fst_occ)
			# May also need to change end date
			if 'DTEND' in ev:
				d = fst_occ - e_inf.start_dt
				end = ev['DTEND'].dt + d
				del(ev['DTEND'])
				ev.add('DTEND', end)


	@staticmethod
	def _event_set_status_from_info(ev:iEvent, e_inf:EntryInfo) -> None:
		# Set event status (cancelled, tentative etc.) from e_inf.
		# Only allow known values.
		if 'STATUS' in ev:
			del(ev['STATUS'])
		if e_inf.status in ('TENTATIVE','CONFIRMED','CANCELLED'):
			ev.add('STATUS', e_inf.status)


	@staticmethod
	def _event_set_location_from_info(ev:iEvent, e_inf:EntryInfo) -> None:
		# Set event location (text string) from e_inf.
		if 'LOCATION' in ev:
			del(ev['LOCATION'])
		if e_inf.location:
			ev.add('LOCATION', e_inf.location)


	@classmethod
	def delete_entry(cls, event:iEvent) -> None:
		# Delete given event.
		# Need to clear cache containing the entry...
		if 'RRULE' in event:
			cls._entry_rep_list = None
		else:
			cls._entry_norep_list_sorted = None
		cls.calConnector.delete_entry(event)


	@classmethod
	def set_toggle_status_entry(cls, event:iEvent, stat:Optional[str]) -> None:
		# Set event STATUS to stat.
		# If STATUS is set and equals stat, toggle it off.
		if 'STATUS' in event:
			if stat==event['STATUS']:
				stat = None # If on, we toggle it off
			del(event['STATUS'])
		if stat in ('TENTATIVE','CONFIRMED','CANCELLED'):
			event.add('STATUS', stat)
		cls.calConnector.update_entry(event) # Write to store


	@classmethod
	def _update_entry_norep_list(cls) -> None:
		# Re-build _entry_norep_list_sorted, if it has been cleared (==None)
		if cls._entry_norep_list_sorted is None:
			# Get events with no repeat rule
			evs = cls.calConnector.cal.walk('vEvent')
			cls._entry_norep_list_sorted = [e for e in evs if 'RRULE' not in e]
			cls._entry_norep_list_sorted.sort()


	@classmethod
	def _update_entry_rep_list(cls) -> None:
		# Re-build _entry_rep_list, if it has been cleared (==None)
		# Possible optimisation: sort most -> least frequent
		# (so don't get last one inserting loads into array) 
		if cls._entry_rep_list is None:
			evs = cls.calConnector.cal.walk('vEvent')
			cls._entry_rep_list = [e for e in evs if 'RRULE' in e]


	@classmethod
	def occurrence_list(cls, start:dt_date, stop:dt_date, include_single:bool=True, include_repeated:bool=True) -> list:
		# Return list of occurences in range start <= . < stop.
		# Designed to be called by View classes to get events in range.
		# An "occurrence" is a pair: (event,datetime)
		#  for repeating entries, datetime may not be the DTSTART entry
		# Needs to also return events that last/end over range??
		ret_list = []
		if include_single:
			cls._update_entry_norep_list()
			# bisect to find starting point
			ii = 0
			llen = len(cls._entry_norep_list_sorted)
			top = llen
			while ii<top:
				mid = (ii+top)//2
				if dt_lt(cls._entry_norep_list_sorted[mid]['DTSTART'].dt,start):
					ii = mid+1
				else:
					top = mid
			# ii is now the start, append occs to ret_list
			while ii < llen:
				e = cls._entry_norep_list_sorted[ii]
				e_st = e['DTSTART'].dt
				if dt_lte(stop, e_st):
					break
				ret_list.append((e,e_st))
				ii += 1
		if include_repeated:
			cls._update_entry_rep_list()
			for e in cls._entry_rep_list:
				merge_repeating_entries_sort(ret_list,e,start,stop)
		return ret_list


#
# Connector class for ICS files
#
class CalendarConnectorICSfile(CalendarConnector):
	BACKUP_PERIOD = 90 # seconds
	BACKUP_EXT = 'bak'
	NEWFILE_EXT = 'new'

	def __init__(self, filename:str):
		self._filename = filename
		if Path(filename).exists():
			with open(filename, 'rb') as file:
				self.cal = iCalendar.from_ical(file.read())
		else: # Create empty calendar file
			self.cal = iCalendar()
			# These aren't added automatically and spec requires them:
			self.cal.add('PRODID', '-//Semiprime//Pygenda//EN')
			self.cal.add('VERSION', '2.0')
		self._backup_saved_time = float('-inf') # so first change creates backup


	def _save_file(self) -> None:
		# Save file to disk/storage. Called after any entry updated.
		# Implementation tries to minimise possibility/extent of data loss.
		file_exists = False
		try:
			mode = os_stat(self._filename).st_mode
			file_exists = True
		except FileNotFoundError:
			mode = stat.S_IRUSR|stat.S_IWUSR # Default - private to user

		# We don't want to just write over current file - it will be
		# zero-length for a time, so a crash would lose data.
		# We save to a new file, so a crash before the write completes
		# will leave original file in place to be opened on restart.
		realfile = os_path.realpath(self._filename)
		tfdir = os_path.dirname(realfile)
		tfpre = '{:s}.{:s}-{:s}-'.format(os_path.basename(realfile),self.NEWFILE_EXT,dt_datetime.now().strftime('%Y%m%d%H%M%S'))
		with tempfile.NamedTemporaryFile(mode='wb', prefix=tfpre, dir=tfdir, delete=False) as tf:
			temp_filename = tf.name
			os_chmod(temp_filename, mode)
			tf.write(self.cal.to_ical())

		# Possibly make a backup of original file before overwriting
		if file_exists and time_monotonic() - self._backup_saved_time > self.BACKUP_PERIOD:
			os_rename(self._filename, '{:s}.{:s}'.format(self._filename,self.BACKUP_EXT))
			self._backup_saved_time = time_monotonic() # re-read time
		# Rename temp saved version to desired name
		os_rename(temp_filename, realfile)


	def add_entry(self, event:iEvent) -> None:
		# Add a new entry component to the file data and write file.
		self.cal.add_component(event)
		self._save_file()


	def update_entry(self, event:iEvent) -> None:
		# event is a component of the file data, so it's already updated.
		# We just need to write the file data.
		self._save_file()


	def delete_entry(self, event:iEvent) -> None:
		# Delete entry component to the file data and write file.
		self.cal.subcomponents.remove(event)
		self._save_file()


#
# Connector class for CalDAV server
# This works by reading all the data from the server on startup, and
# using its copy to calculate query responses. I did it this way
# because, on testing, querying the server directly was too slow:
# using the Radicale server on the Gemini, a simple query with a
# range of a week took more than 3 seconds (0.5 seconds on a laptop).
#
class CalendarConnectorCalDAV(CalendarConnector):
	def __init__(self, url:str, user:str, passwd:str, calname:Optional[str]):
		import caldav # Postponed import, so Pygenda can be used without caldav

		client = caldav.DAVClient(url=url, username=user, password=passwd)
		try:
			principal = client.principal()
		except Exception as excep:
			print('Error: Can\'t connect to CalDAV server at {:s}. Message: {:s}'.format(url,str(excep)), file=stderr)
			raise
		if calname is None:
			calendars = principal.calendars()
			if len(calendars) > 0:
				self.calendar = calendars[0]
			else:
				# Create a calendar with default name
				self.calendar = principal.make_calendar(name='pygenda')
		else:
			# Open or create named calendar
			try:
				self.calendar = principal.calendar(calname)
			except caldav.lib.error.NotFoundError:
				self.calendar = principal.make_calendar(name=calname)

		# Make list of references to events for convenient access
		events = self.calendar.events()
		self.cal = iCalendar()
		for ev in events:
			# Each icalendar_instance is a calendar containing the event.
			# We want to extract the event itself, so walk() & take the first.
			vevent = ev.icalendar_instance.walk('vEvent')[0]
			vevent.__conn_event = ev # Sneakily add ev, for rapid access later
			self.cal.add_component(vevent)


	def add_entry(self, event:iEvent) -> None:
		# Create a new entry component on the server, and locally.
		vcstr = 'BEGIN:VCALENDAR\r\n{:s}END:VCALENDAR\r\n'.format(event.to_ical().decode()) # !! Should we specify encoding for decode()?
		try:
			conn_event = self.calendar.save_event(vcstr)
		except Exception as excep:
			# !! While code is in development, just exit on failure.
			# May change to something "friendlier" later...
			print('Error creating entry on CalDAV server. Message: {:s}'.format(str(excep)), file=stderr)
			exit(-1)

		# Save to local store
		# Add embedded event, so we can modify & save directly
		newevent = conn_event.icalendar_instance.walk('vEvent')[0]
		newevent.__conn_event = conn_event
		self.cal.add_component(newevent)


	def update_entry(self, event:iEvent) -> None:
		# Event struct has been modified, so can just send update to server.
		try:
			event.__conn_event.save() # Write to server
		except Exception as excep:
			# !! While code is in development, just exit on failure.
			# May change to something "friendlier" later...
			print('Error updating entry on CalDAV server. Message: {:s}'.format(str(excep)), file=stderr)
			exit(-1)


	def delete_entry(self, event:iEvent) -> None:
		# Delete entry component from server and in local copy.
		try:
			event.__conn_event.delete() # delete from server
		except Exception as excep:
			# !! While code is in development, just exit on failure.
			# May change to something "friendlier" later...
			print('Error deleting entry on CalDAV server. Message: {:s}'.format(str(excep)), file=stderr)
			exit(-1)
		self.cal.subcomponents.remove(event) # delete local copy


#
# Helper class for repeats_in_range() function (below)
#
class RepeatInfo:
	DAY_ABBR = ('MO','TU','WE','TH','FR','SA','SU')

	def __init__(self, event:iEvent, start:dt_date, stop:dt_date):
		# Note: start argument is INclusive, stop is EXclusive
		rrule = event['RRULE']
		self.start = event['DTSTART'].dt
		# Quickly eliminate some out-of-range cases
		if stop is not None and dt_lte(stop, self.start):
			self.start_in_rng = None
			return
		if start is not None and 'UNTIL' in rrule and dt_lt(rrule['UNTIL'][0],start):
			self.start_in_rng = None
			return

		dt_st = event['DTSTART'].dt
		if rrule['FREQ'][0]=='MONTHLY' and dt_st.day>28:
			raise ValueError('Unsupported MONTHLY (day>28) {} in RRULE'.format(dt_st))
		if rrule['FREQ'][0]=='YEARLY' and dt_st.month==2 and dt_st.day==29:
			raise ValueError('Unsupported YEARLY (29/2) {} in RRULE'.format(dt_st))

		self._set_freq(rrule)
		if 'EXDATE' in event:
			self._set_exdates(event['EXDATE'])
		else:
			self.exdates = None
		self._set_start_in_rng(start)
		self._set_stop(rrule, stop)


	def _set_freq(self, rrule:vRecur) -> None:
		# Set repeat frequency (self.delta).
		# Called during construction.
		freq = rrule['FREQ'][0]
		interval = int(rrule['INTERVAL'][0]) if 'INTERVAL' in rrule else 1
		if 'BYDAY' in rrule and rrule['BYDAY'][0] not in self.DAY_ABBR:
			raise ValueError('Unsupported BYDAY {} in RRULE'.format(rrule['BYDAY']))
		if 'BYMONTH' in rrule and len(rrule['BYMONTH'])>1:
			raise ValueError('Unsupported multi-BYMONTH {} in RRULE'.format(rrule['BYMONTH']))
		if 'BYYEARDAY' in rrule:
			raise ValueError('Unsupported BYYEARDAY in RRULE')
		if 'BYMONTHDAY' in rrule:
			if (freq!='YEARLY' or 'BYMONTH' not in rrule):
				raise ValueError('Unsupported BYMONTHDAY in RRULE (not YEARLY/BYMONTH)')
			# If we get here, it's YEARLY/BYMONTH/BYMONTHDAY
			if len(rrule['BYMONTH'])>1:
				raise ValueError('Unsupported BYMONTHDAY, multi-BYMONTH in RRULE')
			if len(rrule['BYMONTHDAY'])>1:
				raise ValueError('Unsupported multi-BYMONTHDAY in RRULE')
			bmd_day = int(rrule['BYMONTHDAY'][0])
			bmd_month = int(rrule['BYMONTH'][0])
			if bmd_day!=self.start.day or bmd_month!=self.start.month:
				raise ValueError('Unsupported YEARLY/BYMONTH/BYMONTHDAY != DTSTART in RRULE')
		if 'BYSETPOS' in rrule:
			raise ValueError('Unsupported BYSETPOS in RRULE')
		if 'BYHOUR' in rrule:
			raise ValueError('Unsupported BYHOUR in RRULE')
		if 'BYMINUTE' in rrule:
			raise ValueError('Unsupported BYMINUTE in RRULE')
		if 'BYSECOND' in rrule:
			raise ValueError('Unsupported BYSECOND in RRULE')
		if 'BYWEEKNO' in rrule:
			raise ValueError('Unsupported BYWEEKNO in RRULE')
		if freq=='YEARLY':
			self._set_yearly(interval)
		elif freq=='MONTHLY':
			self._set_monthly(interval)
		elif freq=='WEEKLY':
			self._set_weekly(rrule, interval)
		elif freq=='DAILY':
			self._set_daily(interval)
		elif freq=='HOURLY':
			self._set_hourly(interval)
		elif freq=='MINUTELY':
			self._set_minutely(interval)
		else: # unrecognised freq - skip entry
			raise ValueError('Unknown FREQ {:s} in RRULE'.format(freq))


	def _set_yearly(self, interval:int) -> None:
		# Called on construction if a simple yearly repeat
		self.delta = relativedelta(years=interval)


	def _set_monthly(self, interval:int) -> None:
		# Called on construction if a simple monthly repeat
		self.delta = relativedelta(months=interval)


	def _set_weekly(self, rrule:vRecur, interval:int) -> None:
		# Called on construction if a weekly repeat.
		# Try to also handle cases where rrule gives multiple days/startweek.
		days_in_week = 0 # bitmask to be created
		diw_from_st = 1<<self.start.weekday()
		if 'BYDAY' in rrule:
			for d in rrule['BYDAY']:
				days_in_week |= 1<<self.DAY_ABBR.index(d)
			if days_in_week and not days_in_week&diw_from_st:
				# ?? Days are listed, but start day is not among them
				# Ambiguous - raise error
				raise ValueError('First day of weekly repeat not included in repeat day')
		if days_in_week==0 or days_in_week == diw_from_st:
			# Just one day, so a simple delta will work.
			self.delta = timedelta(days=7*interval)
			return

		# Get start of week if present, otherwise default is Monday
		# Don't bother if interval == 1 since makes no difference
		st_wk = 1
		if interval>1 and 'WKST' in rrule:
			st_wk <<= self.DAY_ABBR.index(rrule['WKST'][0])

		# Accumulte list of deltas to add to loop through days
		self.delta = []
		diw_bit = diw_from_st
		i = 0
		while True:
			i += 1
			diw_bit = (diw_bit<<1)%127 # rotate bits left
			if diw_bit==st_wk:
				i+=7*(interval-1)
			if diw_bit & days_in_week:
				self.delta.append(timedelta(days=i))
				i = 0
			if diw_bit==diw_from_st:
				break


	def _set_daily(self, interval:int) -> None:
		# Called on construction if a simple daily repeat
		self.delta = timedelta(days=interval)


	def _set_hourly(self, interval:int) -> None:
		# Called on construction if a simple hourly repeat
		self.delta = timedelta(hours=interval)


	def _set_minutely(self, interval:int) -> None:
		# Called on construction if a simple minutely repeat
		self.delta = timedelta(minutes=interval)


	def _set_exdates(self, exdate) -> None:
		# Initialise object's exdates variable, given an exdate structure.
		# exdate argument is in format of exdate from an ical event, it
		# might be a list, or might not...
		if isinstance(exdate,list):
			l = [dl.dts for dl in exdate]
			self.exdates = sorted(set([datetime_to_date(t[i].dt) for t in l for i in range(len(t))]))
		else:
			l = exdate.dts
			self.exdates = sorted(set([datetime_to_date(t.dt) for t in l]))


	def _set_start_in_rng(self, start:dt_date) -> None:
		# Set start date within given range (that is, on/after 'start')
		self.start_in_rng = self.start
		if isinstance(self.delta,list):
			self.delta_index = 0
		if start is not None:
			# We try to jump to first entry in range
			# First compute d, distance from the range
			d = start - datetime_to_date(self.start)
			if isinstance(self.delta,list):
				if d>timedelta(0): # start provided was after first repeat, so inc
					# Want to do as much as possible in one increment
					per = reduce(lambda x,y:x+y,self.delta) # sum of delta
					s = d//per
					self.start_in_rng += per * s
				# After approximate jump, clear any extras.
				# Do this even if d<=0 to catch case when first dates excluded
				i = 0
				while dt_lt(self.start_in_rng,start) or self.is_exdate(self.start_in_rng):
					self.start_in_rng += self.delta[i]
					i = (i+1)%len(self.delta)
				self.delta_index = i
			else: # single delta, could be timedelta or relativedelta
				if d>timedelta(0): # start provided was after first repeat, so inc
					# Want to do as much as possible in one increment
					if isinstance(self.delta,timedelta):
						s = round(d/self.delta)
					else:
						# self.delta is a relativedelta
						sst = self.start
						try:
							# If we have tzinfo, get rid to make relativedelta
							sst = sst.replace(tzinfo=None)
						except TypeError:
							pass
						d = relativedelta(start, sst)
						if self.delta.years>0:
							s = round(d.years/self.delta.years)
						elif self.delta.months>0:
							s = round((d.years*12 + d.months)/self.delta.months)
						else:
							s = 1
					self.start_in_rng += self.delta * s
				# After approximate jump, clear any extras.
				# Do this even if d<=0 to catch case when first dates excluded
				while dt_lt(self.start_in_rng,start) or self.is_exdate(self.start_in_rng):
					self.start_in_rng += self.delta


	def _set_stop(self, rrule:vRecur, stop:dt_date) -> None:
		# Set stop date in range (i.e. before 'stop' parameter).
		# Note, 'stop' parameter is exclusive (this is more usual)
		# but 'UNTIL' field in iCal is inclusive (according to standard).
		# Internally RepeatInfo class will use an exclusive stop, so:
		#   - Name it stop_exc for clarity
		#   - Take care when using 'UNTIL'
		until = rrule['UNTIL'][0] if 'UNTIL' in rrule else None
		if until is None or (stop is not None and dt_lte(stop, until)):
			self.stop_exc = stop
		elif isinstance(until, dt_datetime):
			self.stop_exc = until+timedelta(milliseconds=1)
		else:
			# dt_datetime is a date only
			self.stop_exc = dt_datetime.combine(until,dt_time(microsecond=1))
		count = rrule['COUNT'][0] if 'COUNT' in rrule else None
		if count is not None:
			if self.exdates is not None:
				raise ValueError('Unsupported COUNT & EXDATE') # !! fix me
			if isinstance(self.delta, list):
				di,md = divmod(count-1, len(self.delta))
				last_by_count = self.start
				last_by_count += di*reduce(lambda x,y:x+y,self.delta) # sum()
				if md:
					last_by_count += reduce(lambda x,y:x+y,self.delta[:md])
			else:
				last_by_count = self.start + (self.delta*(count-1))
			if self.stop_exc is None or dt_lt(last_by_count,self.stop_exc):
				self.stop_exc = last_by_count+timedelta(milliseconds=1) if isinstance(last_by_count,dt_datetime) else dt_datetime.combine(last_by_count,dt_time(microsecond=1))
		if self.stop_exc is None:
			raise RuntimeError('Unbounded repeats will lead to infinite list')


	def is_exdate(self, dt:dt_date) -> bool:
		# Returns True if dt is in exdates, False otherwise.
		# Special case: if dt is a datetime, returns True if date
		# component is in exdates.
		if self.exdates is None:
			return False
		if isinstance(dt,dt_datetime) and dt.date() in self.exdates:
			return True
		return dt in self.exdates


	def __iter__(self) -> 'RepeatIter_simpledelta':
		# Return an iterator for this RepeatInfo
		if self.start_in_rng is not None and isinstance(self.delta, list):
			return RepeatIter_multidelta(self)
		return RepeatIter_simpledelta(self)


class RepeatIter_simpledelta:
	# Iterator class for RepeatInfo where we can just use a simple delta.

	def __init__(self, rinfo:RepeatInfo):
		self.rinfo = rinfo
		self.dt = rinfo.start_in_rng

	def __iter__(self) -> 'RepeatIter_simpledelta':
		# Standard method for iterators
		return self

	def __next__(self) -> dt_date:
		# Return date/dattime for next occurence in range.
		# Excluded dates are taken into account.
		# Raises StopIteration at end of occurence list.
		if self.dt is None or dt_lte(self.rinfo.stop_exc,self.dt):
			raise StopIteration
		r = self.dt
		while True:
			self.dt += self.rinfo.delta
			if not self.rinfo.is_exdate(self.dt):
				break
		return r


class RepeatIter_multidelta(RepeatIter_simpledelta):
	# Iterator class for RepeatInfo where we have different deltas
	# (e.g. we may have different gaps between occurences, such as
	# a weekly repeat that occurs on Mondays and Wednesdays).

	def __init__(self, rinfo:RepeatInfo):
		super().__init__(rinfo)
		self.i = rinfo.delta_index

	def __next__(self) -> dt_date:
		# Return date/dattime for next occurence in range.
		# Excluded dates are taken into account.
		# Raises StopIteration at end of occurence list.
		if self.dt is None or dt_lte(self.rinfo.stop_exc,self.dt):
			raise StopIteration
		r = self.dt
		while True:
			self.dt += self.rinfo.delta[self.i]
			self.i = (self.i+1)%len(self.rinfo.delta)
			if not self.rinfo.is_exdate(self.dt):
				break
		return r


def merge_repeating_entries_sort(target:list, ev:iEvent, start:dt_date, stop:dt_date) -> None:
	# Given a sorted list of occurrences, 'target', and a single
	# repeating event 'ev', splice the repeats of ev from 'start'
	# to 'stop' into 'target', keeping it sorted.
	ev_reps = repeats_in_range(ev, start, stop)
	i,j = 0,0
	end_i = len(target)
	end_j = len(ev_reps)
	while i<end_i and j<end_j:
		if dt_lt(ev_reps[j],target[i][1]):
			target.insert(i,(ev,ev_reps[j]))
			j += 1
			end_i +=1
		i += 1
	if j<end_j:
		target.extend([(ev,dt) for dt in ev_reps[j:]])


def first_occ(rrstr:str, dtstart:dt_date) -> dt_date:
	# Returns the date or datetime of the first occurrence of an event,
	# given an rrule and an (earliest possible) start date.
	# Does not take account of excluded dates.
	rr = rrulestr(rrstr, dtstart=dtstart)
	has_time = isinstance(dtstart, dt_datetime)
	st = dtstart if has_time else dt_datetime.combine(dtstart, dt_time())
	ret = rr.after(st, inc=True)
	if ret and not has_time:
		ret = ret.date()
	return ret


def repeats_in_range_with_rrstr(ev:iEvent, start:dt_date, stop:dt_date) -> list:
	# Get repeat dates within given range using dateutil.rrule module.
	# Slow, but comprehensive. Used as a fallback from repeats_in_range()
	# when quick methods can't be used.
	# Repeats are super clunky.
	# Can caching results help?
	rrstr = ev['RRULE'].to_ical().decode('utf-8')
	dt = ev['DTSTART'].dt
	hastime = isinstance(dt,dt_datetime)
	exd = 'EXDATE' in ev
	rr = rrulestr(rrstr,dtstart=dt,forceset=exd)
	st = dt_datetime.combine(start, dt_time()) # set time to midnight
	sp = dt_datetime.combine(stop, dt_time())
	if hastime:
		# Could add tzinfo to combine() calls above, but this allows Python<3.6
		st = st.replace(tzinfo=dt.tzinfo)
		sp = sp.replace(tzinfo=dt.tzinfo)
	sp -= timedelta(milliseconds=1)
	ret = rr.between(after=st,before=sp,inc=True)
	if not hastime:
		ret = [d.date() for d in ret]
	if exd:
		exdate_list = [ev['EXDATE'][i].dts[j] for i in range(len(ev['EXDATE'])) for j in range(len(ev['EXDATE'][i].dts))] if isinstance(ev['EXDATE'], list) else ev['EXDATE'].dts
		if hastime:
			for de in exdate_list:
				if isinstance(de.dt,dt_datetime):
					ret = [d for d in ret if d!=de.dt]
				else:
					ret = [d for d in ret if d.date()!=de.dt]
		else:
			for de in exdate_list:
				if not isinstance(de.dt,dt_datetime):
					ret = [d for d in ret if d!=de.dt]
	return ret


def repeats_in_range(ev:iEvent, start:dt_date, stop:dt_date) -> list:
	# Given a repeating event ev, return list of occurrences from
	# dates start to stop.
	try:
		r_info = RepeatInfo(ev, start, stop)
	except ValueError as err:
		# RepeatInfo doesn't handle this type of repeat.
		# Fall back to using rrule - more complete, but slower for simple repeats
		print('Notice: Fallback to unoptimised repeat for "{:s}" ({:s})'.format(ev['SUMMARY'],str(err)), file=stderr)
		return repeats_in_range_with_rrstr(ev, start, stop)
	ret = list(iter(r_info))
	# Uncomment the next two lines to test calculated values (slow!)
	#if ret != repeats_in_range_with_rrstr(ev, start, stop):
	#	print('Error: Wrong repeats for "{:s}"'.format(ev['SUMMARY']), file=stderr)
	return ret
